/*****************************************************************************
  ColorImage.cc contains implementations of the functions of the EdgeImage
  class declared in ColorImage.h
  Programmer: Kirill Yevseyev
  Last Updated: 8-18-94
  $Log: ColorImage.cc,v $
// Revision 1.8  1995/01/06  19:08:46  heller
// rint and bzero
//
// Revision 1.7  1994/11/21  13:12:55  heller
// Replaced fprintf(stderr,...) with cerr << ...
//
// Revision 1.6  1994/11/16  17:32:53  heller
// Updated NAN handling
//
// Revision 1.5  1994/10/31  15:44:28  heller
// Fix NaN handling (undefined Hue value).
//
 *****************************************************************************/

static char rscid[] = "$Id: ColorImage.cc,v 1.8 1995/01/06 19:08:46 heller Exp $";

#include <ColorImage.h>
#include <NewPlane.h>
#include <iostream.h>
#include <Isrstream.h>

#define UNDEFINED (Token::NaN_Value.d)
#define RAD_TO_DEG 180 / M_PI
#define MAX(a, b)  ((a) > (b) ? (a) : (b))
#define MIN(a, b)  ((a) < (b) ? (a) : (b))

#define bzero(p,n) memset(p,0,n)

ColorImage::ColorImage ()
/* Default constructor */
{
  red_plane = NULL;
  green_plane = NULL;
  blue_plane = NULL;
  hue_plane = NULL;
  saturation_plane = NULL;
  value_plane = NULL;
  intensity_plane = NULL;
  map_plane = NULL;
}

ColorImage::ColorImage (int XSize, int YSize, int XOrigin, int YOrigin,
              AxisDirection XDir, AxisDirection YDir, Boolean RGB)
{
  if (RGB)
  {
    red_plane = new BytePlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Red");
    green_plane = new BytePlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Green");
    blue_plane = new BytePlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Blue");
    hue_plane = NULL;
    saturation_plane = NULL;
    value_plane = NULL;
  } else
  {
    hue_plane = new FloatPlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Hue");
    saturation_plane = new FloatPlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Saturation");
    value_plane = new FloatPlane(XSize,YSize,XOrigin,YOrigin,XDir,YDir,"Value");
    red_plane = NULL;
    green_plane = NULL;
    blue_plane = NULL;
  }
  intensity_plane = NULL;
  map_plane = NULL;
}

/*************************************************************/

static FloatPlane *copy_color_plane (FloatPlane *inplane, char *planename)
/* Local function for copying planes. Returns a pointer to new FloatPlane */
{
  int minx, maxx, miny, maxy;
  FloatPlane *outplane = new FloatPlane (*inplane, planename);
  inplane->GetMinMax (minx, maxx, miny, maxy);
  for (int i = minx; i <= maxx; i ++)
    for (int j = miny; j <= maxy; j ++)
      outplane->pixel (i, j) = inplane->pixel (i, j);
  return outplane;
}

/*************************************************************/

static BytePlane *copy_color_plane (BytePlane *inplane, char *planename)
/* Local function for copying planes. Returns a pointer to new BytePlane */
{
  int minx, maxx, miny, maxy;
  BytePlane *outplane = new BytePlane (*inplane, planename);
  inplane->GetMinMax (minx, maxx, miny, maxy);
  for (int i = minx; i <= maxx; i ++)
    for (int j = miny; j <= maxy; j ++)
      outplane->pixel (i, j) = inplane->pixel (i, j);
  return outplane;
}

/*************************************************************/

ColorImage::ColorImage (BytePlane *red, BytePlane *green, BytePlane *blue)
/* Constructor */
{
  red_plane = copy_color_plane (red, "red");
  green_plane = copy_color_plane (green, "green");
  blue_plane = copy_color_plane (blue, "blue");
  map_plane = NULL;
  intensity_plane = NULL;
  hue_plane = NULL;
  saturation_plane = NULL;
  value_plane = NULL;
}

/*************************************************************/

ColorImage::ColorImage (FloatPlane *hue, FloatPlane *saturation, 
			FloatPlane *value)
/* Constructor */
{
  hue_plane = copy_color_plane (hue, "hue");
  saturation_plane = copy_color_plane (saturation, "saturation");
  value_plane = copy_color_plane (value, "value");
  map_plane = NULL;
  intensity_plane = NULL;
  red_plane = NULL;
  green_plane = NULL;
  blue_plane = NULL;
}    

/*************************************************************/

ColorImage::~ColorImage ()
/* Destructor */
{
  if (red_plane != NULL)
    delete red_plane;
  if (green_plane != NULL)
    delete green_plane;
  if (blue_plane != NULL)
    delete blue_plane;
  if (hue_plane != NULL)
    delete hue_plane;
  if (saturation_plane != NULL)
    delete saturation_plane;
  if (value_plane != NULL) 
    delete value_plane;
  if (intensity_plane != NULL)
    delete intensity_plane;
  if (map_plane != NULL)
    delete map_plane;
}

/*************************************************************/

void ColorImage::input (Isristream& is)
{
  is >> "red = ";
  red_plane = (BytePlane *) read_token_ptr (is);
  is >> "green = ";
  green_plane = (BytePlane *) read_token_ptr (is);
  is >> "blue = ";
  blue_plane = (BytePlane *) read_token_ptr (is);
  is >> "hue = ";
  hue_plane = (FloatPlane *) read_token_ptr (is);
  is >> "saturation = ";
  saturation_plane = (FloatPlane *) read_token_ptr (is);
  is >> "value = ";
  value_plane = (FloatPlane *) read_token_ptr (is);
  is >> "intensity = ";
  intensity_plane = (BytePlane *) read_token_ptr (is);
}

/*************************************************************/

void ColorImage::output(Isrostream& os)
{
  os << "red = " << red_plane << "\n";
  os << "green = " << green_plane << "\n";
  os << "blue = " << blue_plane << "\n";
  os << "hue = " << hue_plane << "\n";
  os << "saturation = " << saturation_plane << "\n";
  os << "value = " << value_plane << "\n";
  os << "intensity = " << intensity_plane << "\n";
}

/*************************************************************/

void ColorImage::trace ()
{
  Register (red_plane);
  Register (green_plane);
  Register (blue_plane);

  Register (hue_plane);
  Register (saturation_plane);
  Register (value_plane);

  Register (intensity_plane);
}

/*************************************************************/

void ColorImage::BitDraw (Display* display, Drawable drawable, GC gc,
                         Transform2D* trans, float xmin, float ymin,
                         float xmax, float ymax, Visual *visual,
                         unsigned int depth, Colormap cmap)
/* Draws a ColorImage */
{
  if (depth < 8 ||
      visual->c_class == StaticGray || 
      visual->c_class == GrayScale)
  {
    if (intensity_plane == NULL)
      compute_intensity ();
    if (intensity_plane == NULL)
      return;
    intensity_plane->BitDraw (display, drawable, gc, trans, xmin, ymin,
                     xmax, ymax, visual, depth, cmap);
  } else {
    if (map_plane == NULL)
    {
    	MakeMapPlane(128);
    }
    double min,max;
    map_plane->GetMinMaxVal(min,max);
    int nc = (int) (max + 1);
    map_plane->SetColorMap(rmap,gmap,bmap,nc,display,cmap);
    map_plane->BitDraw (display, drawable, gc, trans, xmin, ymin,
                     xmax, ymax, visual, depth, cmap);
  }
}

/*************************************************************/

static unsigned char float_pixel_to_byte (float v)
{
  if (v > 255) return 255;
  else if (v < 0) return 0;
  return (unsigned char) (v+0.5);
}

/*************************************************************/

int ColorImage::compute_red_green_blue ()
/* Computes red, green, and blue planes if hue, saturation, and value 
   are known. */
{
  if ((hue_plane == NULL) || (saturation_plane == NULL) || 
      (value_plane == NULL))
    return -1;
 
  int minx, maxx, miny, maxy;
  BytePlane *red_pl;
  red_pl = new BytePlane (*hue_plane, "red");
  BytePlane *green_pl;
  green_pl = new BytePlane (*hue_plane, "green");
  BytePlane *blue_pl;
  blue_pl = new BytePlane (*hue_plane, "blue");
  
  value_plane->GetMinMax (minx, maxx, miny, maxy);
  for (int i = minx; i <= maxx; i ++)
    for (int j = miny; j <= maxy; j ++) {

      float h = hue_plane->pixel (i, j);
      float s = saturation_plane->pixel (i, j);
      float v = value_plane->pixel (i, j);

      if (s == 0) 
      {
      	double hd = h;
      	if (NaN(hd))
      	{
	  red_pl->pixel (i, j) = float_pixel_to_byte (v);
	  green_pl->pixel (i, j) = float_pixel_to_byte (v);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (v);
	} else
	{
	  cerr << "Saturation is zero and hue has a value: " << h << endl;
	  return -1;
	}
      }
      else {
	h = h * RAD_TO_DEG;
	if (h == 360)
	  h = 0;
	h = h / 60;
	/* hue is now in [0, 6) */
	int k = (int) floor (h);
	/* largest integer <= hue */
	float f = h - k;
	/* fractional part of hue */
	float p = v * (1 - s);
	float q = v * (1 - (s * f));
	float t = v * (1 - (s * (1 - f)));
	switch (k) {
	case 0:
	  red_pl->pixel (i, j) = float_pixel_to_byte (v);
	  green_pl->pixel (i, j) = float_pixel_to_byte (t);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (p);
	  break;
	case 1:
	  red_pl->pixel (i, j) = float_pixel_to_byte (q);
	  green_pl->pixel (i, j) = float_pixel_to_byte (v);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (p);
	  break;
	case 2:
	  red_pl->pixel (i, j) = float_pixel_to_byte (p);
	  green_pl->pixel (i, j) = float_pixel_to_byte (v);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (t);
	  break;
	case 3:
	  red_pl->pixel (i, j) = float_pixel_to_byte (p);
	  green_pl->pixel (i, j) = float_pixel_to_byte (q);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (v);
	  break;
	case 4:
	  red_pl->pixel (i, j) = float_pixel_to_byte (t);
	  green_pl->pixel (i, j) = float_pixel_to_byte (p);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (v);
	  break;
	case 5:
	  red_pl->pixel (i, j) = float_pixel_to_byte (v);
	  green_pl->pixel (i, j) = float_pixel_to_byte (p);
	  blue_pl->pixel (i, j) = float_pixel_to_byte (q);
	  break;
	}
      }
    }
  red_plane = red_pl;
  green_plane = green_pl;
  blue_plane = blue_pl;
  return 1;
}

/*************************************************************/

float Max (float a, float b, float c)
/* Local function that finds the maximum of 3 parameters. */
{
  return MAX (MAX (a, b), MAX (b, c)); 
}

/*************************************************************/

float Min (float a, float b, float c)
/* Local function that finds the maximum of 3 parameters. */
{
  return MIN (MIN (a, b), MIN (b, c)); 
}

/*************************************************************/
 
int ColorImage::compute_hue_saturation_value ()
/* Computes hue, saturation, and value planes if red, green, and blue
   are known. */
{
  if ((red_plane == NULL) || (green_plane == NULL) || (blue_plane == NULL))
    return -1;
  
  int minx, maxx, miny, maxy;
  float max, min, gc, rc, bc;
  FloatPlane *hue_pl;
  hue_pl = new FloatPlane (*red_plane, "hue");
  FloatPlane *saturation_pl;
  saturation_pl = new FloatPlane (*red_plane, "saturation");
  FloatPlane *value_pl;
  value_pl = new FloatPlane (*red_plane, "value");

  red_plane->GetMinMax (minx, maxx, miny, maxy);
  for (int i = minx; i <= maxx; i ++)
    for (int j = miny; j <= maxy; j ++) {

      float r = (float) red_plane->pixel (i, j);
      float g = (float) green_plane->pixel (i, j);
      float b = (float) blue_plane->pixel (i, j);

      max = Max (r, g, b);
      min = Min (r, g, b);

      value_pl->pixel (i, j) = max;
      saturation_pl->pixel (i, j) = (max != 0 ? ((max - min) / max) : 0);
      if (saturation_pl->pixel (i, j) == 0)
	hue_pl->pixel (i, j) = UNDEFINED;
      else {
	rc = (max - r) / (max - min);
	/* rc = "distance" of color from red */
	gc = (max - g) / (max - min);
	bc = (max - b) / (max - min);
	if (r == max)
	  hue_pl->pixel (i, j) = bc - gc; 
	/* resulting color between yellow and magenta */
	else if (g == max)
	  hue_pl->pixel (i, j) = 2 + rc - bc;
	/* resulting color between cyan and yellow */
	else if (b == max)
	  hue_pl->pixel (i, j) = 4 + gc - rc;
	/* resulting color between magenta and cyan */
	float h = hue_pl->pixel (i, j);
	if (h < 0)
	  hue_pl->pixel (i, j) = h + 2 * M_PI;
	  /*make nonnegative */
      }
    }
  hue_plane = hue_pl;
  saturation_plane = saturation_pl;
  value_plane = value_pl;
  return 1;
}

/*************************************************************/

void ColorImage::compute_intensity ()
/* Computes intensity_plane. */
{
  int minx, maxx, miny, maxy;
  BytePlane *intensity_pl;
  intensity_pl = new BytePlane (*red_plane, "intensity");

  red_plane->GetMinMax (minx, maxx, miny, maxy);
  for (int i = minx; i <= maxx; i ++)
    for (int j = miny; j <= maxy; j ++) 
      intensity_pl->pixel (i, j) = (red ()->pixel (i, j) + 
				    green ()->pixel (i, j) +
				    blue ()->pixel (i, j)) / 3;
  intensity_plane = intensity_pl;
}

/*************************************************************/

void ColorImage::SetRedGreenBlue (BytePlane *r, BytePlane *g, BytePlane *b)
/* Sets red, green, and blue to passed parameters. */
{
  if (red_plane != NULL) delete red_plane;
  red_plane = copy_color_plane (r, "red");
  if (green_plane != NULL) delete green_plane;
  green_plane = copy_color_plane (g, "green");
  if (blue_plane != NULL) delete blue_plane;
  blue_plane = copy_color_plane (b, "blue"); 

  if (hue_plane != NULL)
    delete hue_plane;
  hue_plane = NULL;
  if (saturation_plane != NULL)
    delete saturation_plane;
  saturation_plane = NULL;
  if (value_plane != NULL) 
    delete value_plane;
  value_plane = NULL;
  if (intensity_plane != NULL)
    delete intensity_plane;
  intensity_plane = NULL;
  compute_intensity();
  if (map_plane != NULL)
    delete map_plane;
  map_plane = NULL;

}

/*************************************************************/

void ColorImage::SetHueSaturationValue (FloatPlane *h, FloatPlane *s, 
					FloatPlane *v)
/* Sets hue, saturation, value to passed parameters. */
{
  if (hue_plane != NULL) delete hue_plane;
  hue_plane = copy_color_plane (h, "hue");
  if (saturation_plane != NULL) delete saturation_plane;
  saturation_plane= copy_color_plane (s, "saturation");
  if (value_plane != NULL) delete value_plane;
  value_plane = copy_color_plane (v, "value");
  if (red_plane != NULL)
    delete red_plane;
  red_plane = NULL;
  if (green_plane != NULL)
    delete green_plane;
  green_plane = NULL;
  if (blue_plane != NULL) 
    delete blue_plane;
  blue_plane = NULL;
  if (intensity_plane != NULL)
    delete intensity_plane;
  intensity_plane = NULL;
  compute_intensity();
  if (map_plane != NULL)
    delete map_plane;
  map_plane = NULL;
}

/*************************************************************/

/* The following function is the hash function for indexing into the
   list of color caches */
#define f(r,g,b) (chopt[r] << 2*nchbits) | (chopt[g] << nchbits) | chopt[b]

/* Note: MAX_CHANLS must be of the form 2^(3*n), n an integer! */
#define MAX_CHANLS 262144                 /* Max number of hash indices */
#define MAX_CHANL_BITS 18                 /* Log base 2 of MAX_CHANLS */
static int nchanls;                              /* Number of hash indices */
static int nchbits;                              /* Number of bits USED for hash, per band */
static ColorImage::color *chead,*ctail;               /* Composite color list */
static ColorImage::color **ch,**ct;                   /* Hash index color lists */

static ColorImage::box *boxhead,*boxtail;

static ColorImage::pixel *pixpile,*pixels;            /* Pointers to pixel area */
#define pmalloc() pixpile++               /* Pixel handout function */

#define MAX_CBLOCKS 2048                  /* Max number of color blocks */
#define CNUM 2048                         /* Number of colors/block */
static ColorImage::color *cblock[MAX_CBLOCKS];        /* Color block pointer array */
static int cbcount,ccount;                       /* Color block count, colors/block*/

static int hash;                                 /* Global hash index */
static unsigned char rshift[256];                /* Shift tables */
static unsigned char lshift[256];                /* Shift tables */
static unsigned char chopt[256];                 /* Chop tables */

void ColorImage::MakeMapPlane(int ncolors,int bits,double fraction)
{
    int shift;                            /* Shift distance for precision */
    int chop;                             /* Shift distance for hash size */
    int i;
    int XS,YS;

    /* Initialize variables used for fast dynamic memory handouts */
    cbcount = 0;
    ccount = 0;
    red()->GetSize(XS,YS);
    pixels = new pixel[XS*YS];
    if (pixels == NULL)
      {
      	cerr << "ColorImage::MakeMapPlane: cannot allocate pixel space!\n";
        return;
      }
    pixpile = pixels;

    /* Build shift and chop tables */
    nchbits = bits;
    if (nchbits > MAX_CHANL_BITS/3) nchbits = MAX_CHANL_BITS/3;
    shift = 8-bits;                   /* Get shift distance */
    chop  = bits-nchbits;             /* Get chop distance */
    if (chop < 0) chop = 0;           /* Force lower bound */
    nchanls = (int) pow((double)2.0,(double)(3.0*nchbits)); /* Compute  number of indices */
    for (i=0; i<256; i++)
      {
        lshift[i] = i << shift;
        rshift[i] = i >> shift;
        chopt[i] = i >> chop;
      }

    load_color_list();            /* Make color list */
    compress_colors(&ncolors,fraction); /* Squeeze the color list */
    gamut_adjust_image(ncolors);       /* Make squeezed image */
    free_colors();                     /* Give back all memory */

  }

void ColorImage::load_color_list()                 /* Build the color list */
  {
    register int clr;
    register color *pc;
    register unsigned char *p1,*p2,*p3;
    register int i,j;
    unsigned char r,g,b;
    int XSize,YSize;

    chead = NULL;                         /* Initialize color list */
    ctail = NULL;

    /* Allocate space for hash index lists */
    
    ch = new color* [nchanls];
    ct = new color* [nchanls];
    if (ch == NULL || ct == NULL)
      {
	cerr << "lvgamut: Unable to allocate hash lists!\n";
        exit(1);
      }
    bzero(ch,nchanls*sizeof(struct color *));
    bzero(ct,nchanls*sizeof(struct color *));

    p1 = red()->GetDataPtr();
    p2 = green()->GetDataPtr();
    p3 = blue()->GetDataPtr();

    red()->GetSize(XSize,YSize);
    for (j=0; j<YSize; j++)
      {
        for (i=0; i<XSize; i++)
          {
            r = rshift[(int)(*p1++)];                       /* Quantize */
            g = rshift[(int)(*p2++)];
            b = rshift[(int)(*p3++)];
            /* Go update the clist */
            clr = ((int)r << 16) | ((int)g << 8) | (int)b;
            hash = f(r,g,b);
            pc = ch[hash];
            while (pc != NULL)
              {
                if (pc->clr == clr) break;
                else pc = pc->next;
              }     
            if (pc == NULL)
              {
                /* Add a color with the current values */
                pc = cmalloc();
                if (ch[hash] == NULL && ct[hash] == NULL) /* Empty color list */
                  {
                    ch[hash] = pc;
                    ct[hash] = pc;
                    pc->prev = NULL;
                    pc->next = NULL;
                  }
                else                                /* Not empty, add to head */
                  {
                    pc->prev = NULL;
                    pc->next = ch[hash];
                    ch[hash]->prev = pc;
                    ch[hash] = pc;
                  }
                pc->r = r;
                pc->g = g;
                pc->b = b;
                pc->clr = clr;
                pc->count = 1;
                pc->pixlist = pmalloc();
                pc->pixlist->next = NULL;
/* The following line replaces the next two */
                pc->pixlist->offset = j*XSize+i;
/*
                pc->pixlist->x = i;
                pc->pixlist->y = j;
*/
                pc->pixtail = pc->pixlist;
              }
            else
              {
                /* Update the pixel list of this color */
                pc->count++;
                pc->pixtail->next = pmalloc();
                pc->pixtail = pc->pixtail->next;
                pc->pixtail->next = NULL;
/* The following line replaces the next two */
                pc->pixtail->offset = j*XSize+i;
/*
                pc->pixtail->x = i;
                pc->pixtail->y = j;
*/
                if (pc != ch[hash]) gamut_head_swap(pc);
              }
          }
      }

    /* Now append all color lists together */
    for (i=0; i<nchanls; i++)
      {
        if (ch[i] != NULL)     /* If have something to add */
          {
            if (ctail != NULL) /* If main list not empty */
              {
                ctail->next = ch[i];
                ch[i]->prev = ctail;
                ctail = ct[i];
              }
            else               /* Main list is empty */
              {
                chead = ch[i];
                ctail = ct[i];
              }
          }
      }
    delete ch;
    delete ct;

  }
   
ColorImage::color *ColorImage::cmalloc()
  {
    /* Maintain a list of blocks of CNUM colors, allocated in chunks and then
       doled out as needed. */
    if (cbcount == 0 || ccount >= CNUM)
      {
        if (cbcount >= MAX_CBLOCKS)
          {
	    cerr << "ColorImage::cmalloc: too many color blocks!\n";
	    cerr << "                     recompile with larger MAX_CBLOCKS\n";
            exit(1);
          }
        cblock[cbcount]=(struct color *)malloc(CNUM*sizeof(struct color));
        if (cblock[cbcount] == NULL)
          {
            cerr << "ColorImage::cmalloc: unable to allocate color block!\n";
            exit(1);
          }
        cbcount++;
        ccount = 0;
      }
    return(cblock[cbcount-1]+(ccount++));
  }

void ColorImage::gamut_head_swap(register color *p)
  {
    register int k;

    k = f(p->r,p->g,p->b);
    /* Delete current color from list */
    if (p == ct[k])
      {
        p->prev->next = p->next;
        ct[k] = p->prev;
      }
    else
      {
        p->prev->next = p->next;
        p->next->prev = p->prev;
      }

    /* Re-insert color at head of list*/
    p->next = ch[k];
    p->prev = NULL;
    ch[k]->prev = p;
    ch[k] = p;
  }
 
void ColorImage::compress_colors(int *ncolors,float fraction)
  {
    int i,n,count,popok;
    unsigned char r,g,b,cmax;
    color *p1,*p2;
    register box *b1,*b2;

    /* Go find out how many colors are present in the image */
    count = 0;
    p1 = chead;
    while (p1 != NULL)
      {
        count++;
        p1 = p1->next;
      }

    /* Now see if the user has requested more colors than exist in the
       image. Spit out a message if verbose is true. Return only the
       number of colors that actually exist. */
    if (*ncolors > count)
      {
        *ncolors = count;
      }

    /* Initialize box structure */
    boxhead = NULL;
    boxtail = NULL;

    /* Make first box and attach all colors to it */
    b1 = new box;
    if (b1 == NULL)
      {
        cerr << "ColorImage::compress_colors: Not enough memory for root box!\n";
        exit(1);
      }
    boxhead = b1;
    boxtail = b1;
    b1->colors = chead;
    gamut_set_box_bounds(b1);
    b1->next = NULL;
    
    /* Now go through and perform a median cut on the box list */
    /* until the desired number of boxes (colors) are formed. */
    popok = TRUE;
    for (i=0; i<(*ncolors)-1; i++)
      {
        /* Hunt down the biggest box */
        if (i < (*ncolors)*(1.0-fraction) && popok == TRUE) 
          {
            b1 = find_biggest_box1();  /* Population */
            if (b1 == NULL)
              {
                cerr << "ColorImage::compress_colors: Largest box population was 0!\n";
                exit(1);
              }
          }
        else 
          {
retry:      b1 = find_biggest_box2();                /* 2-norm */
            if (b1 == NULL)
              {
                cerr << "ColorImage::compress_colors: Largest box 2-norm was 0!\n";
                exit(1);
              }
          }

        /* First make new box */
        b2 = new box;
        if (b2 == NULL)
          {
            cerr << "ColorImage::compress_colors: Not enough memory for additional box!\n";
            exit(1);
          }
        b2->next = NULL;
        b2->colors = NULL;

        /* Copy the color bounds to the new box. */
        b2->rmax = b1->rmax; b2->gmax = b1->gmax; b2->bmax = b1->bmax;
        b2->rmin = b1->rmin; b2->gmin = b1->gmin; b2->bmin = b1->bmin;

        /* Compute boundary for splitting the color space */
        r = b1->rmax-b1->rmin;
        g = b1->gmax-b1->gmin;
        b = b1->bmax-b1->bmin;
        cmax = b; n = 3;                              /* Split on blue */
        if (g > cmax) { cmax = g; n = 2; }            /*    or on green */
        if (r > cmax) { cmax = r; n = 1; }            /*    or on red */

        /* Check to make sure that the axis being split has a legal span */
        if (popok == TRUE) /* Splitting on population */
          {
            switch(n)
              {
                case 1:
                  if (r < 2)
                    {
                      popok = FALSE;
                      free(b2);
                      goto retry;
                    }
                  break;
                case 2:
                  if (g < 2)
                    {
                      popok = FALSE;
                      free(b2);
                      goto retry;
                    }
                  break;
                case 3:
                  if (b < 2)
                    {
                      popok = FALSE;
                      free(b2);
                      goto retry;
                    }
                  break;
              }
          }
        else     /* Splitting on 2-norm */
          {
            switch(n)
              {
                case 1:
                  if (r >= 1) break;
                case 2:
                  if (g >= 1) break;
                case 3:
                  if (b >= 1) break;
                cerr << "ColorImage::compress_colors: Couldn't find a splittable color subspace!\n";
                cerr << "                             Use more precision, or request less colors.\n";
                exit(1);
              }
          }

        /* Update the global boxtail pointer */
        boxtail->next = b2;
        boxtail = boxtail->next;

        /* Modify the color bounds for the original and new boxes */
        switch(n)
          {
            case 1:                              /* Splitting on red */
              b1->rmax = b1->rmin+r/2;
              b2->rmin = b2->rmax-r/2;
              break;
            case 2:                              /* Splitting on blue */
              b1->gmax = b1->gmin+g/2;
              b2->gmin = b2->gmax-g/2;
              break;
            case 3:                              /* Splitting on green */
              b1->bmax = b1->bmin+b/2;
              b2->bmin = b2->bmax-b/2;
              break;
          }

        /* Partition the color space of the first box */
        p1 = b1->colors;           /* Color list in original box */
        p2 = b2->colors;           /* End of color list in new box */
        while (p1 != NULL)
          {
            switch(n)
              {
                case 1:                          /* Splitting on red */
                  if (p1->r > b1->rmax) box_move_color(b1,b2,&p1,&p2);
                  else p1 = p1->next;
                  break;
                case 2:                          /* Splitting on green */
                  if (p1->g > b1->gmax) box_move_color(b1,b2,&p1,&p2);
                  else p1 = p1->next;
                  break;
                case 3:                          /* Splitting on blue */
                  if (p1->b > b1->bmax) box_move_color(b1,b2,&p1,&p2);
                  else p1 = p1->next;
                  break;
              }
          }

        /* Re-compute the color bounds for both boxes */
        gamut_set_box_bounds(b1);
        gamut_set_box_bounds(b2);
      }
  }


void ColorImage::box_move_color(register box *b1,register box *b2,register color **p1,register color **p2)
  {
    /* Move the color p1 from box b1's color list to box b2's color list,
       which has the tail of its color list marked by o2. c1 must me left
       pointing to the next color in b1's color list, which p2 must be left
       pointing at the last color in b2's color list. */

    register color *c1;

    c1 = (*p1)->next;           /* Remember where "next" is. */

    /* Unlink color */
    if (b1->colors == *p1 && (*p1)->next != NULL)    /* p1 is head of color list */
      {
        b1->colors = (*p1)->next;
        (*p1)->next->prev = NULL;
      }
    else if (b1->colors == *p1 && (*p1)->next == NULL) /* p1 is only color */
      {
        b1->colors = NULL;
      }
    else if (b1->colors != *p1 && (*p1)->next == NULL) /* p1 is tail of list */
      {
        (*p1)->prev->next = NULL;
      }
    else
      {
        (*p1)->next->prev = (*p1)->prev;      /* p1 in middle of list */
        (*p1)->prev->next = (*p1)->next;
      }
    (*p1)->next = NULL;
    (*p1)->prev = NULL;

    /* Append the color to the correct place */
    if (b2->colors == NULL)                  /* Empty color list */
       {
         b2->colors = *p1;
         *p2 = *p1;
       }
    else                                     /* Non-empty color list */
      {
        (*p2)->next = *p1;
        (*p1)->prev = *p2;
        *p2 = *p1;
      }

    *p1 = c1;                   /* Regurgitate "next". */
  }

ColorImage::box *ColorImage::find_biggest_box1()
 {
   /* Find the box with the largest population */
   register struct box *b,*b1;
   register int max;

   b = boxhead;
   b1 = NULL;
   max = 0;
   while (b != NULL)
     {
       if (b->count > max)
         {
           max = b->count;
           b1 = b;
         }
       b = b->next;
     }
   return(b1);
 }

ColorImage::box *ColorImage::find_biggest_box2()
 {
   /* Find the box with the largest 2-norm */
   register struct box *b,*b1;
   register int maxnorm,norm,red,green,blue;

   b = boxhead;
   b1 = NULL;
   maxnorm = 0;
   while (b != NULL)
     {
       red = (int)(b->rmax)-(int)(b->rmin);
       green = (int)(b->gmax)-(int)(b->gmin);
       blue = (int)(b->bmax)-(int)(b->bmin);
       norm = red*red+green*green+blue*blue;
       if (norm > maxnorm)
         {
           maxnorm = norm;
           b1 = b;
         }
       b = b->next;
     }
   return(b1);
 }

void ColorImage::gamut_set_box_bounds(box *b)
  {
    register color *c;
    register int rmin,rmax,bmin,bmax,gmin,gmax;
    register int n;

    rmin = gmin = bmin = 255;
    rmax = gmax = bmax = 0;

    n = 0;
    c = b->colors;
    while (c != NULL)
      {
        if ((int)(c->r) > rmax) rmax = (int)(c->r);
        if ((int)(c->r) < rmin) rmin = (int)(c->r);
        if ((int)(c->g) > gmax) gmax = (int)(c->g);
        if ((int)(c->g) < gmin) gmin = (int)(c->g);
        if ((int)(c->b) > bmax) bmax = (int)(c->b);
        if ((int)(c->b) < bmin) bmin = (int)(c->b);
        c = c->next;
        n++;
      }

    b->rmax = rmax; b->rmin = rmin;
    b->gmax = gmax; b->gmin = gmin;
    b->bmax = bmax; b->bmin = bmin;
    b->count = n;
  }

void ColorImage::gamut_adjust_image(int ncolors)
  {
    register unsigned char *d,*rm,*gm,*bm;
    register color *c;
    register pixel *p;
    register int r,g,b;
    register box *bp;
    register int i;
    int XSize,YSize;
    

    red()->GetSize(XSize,YSize);
    /* Allocate space for the new image data, and attach to the header */
    if (map_plane != NULL) delete map_plane;
    map_plane = new BytePlane(*red());
    d = map_plane->GetDataPtr();

    /* Read thru the box list stepping thru its color list and placing
       the pixels in the output image */
    i = 0;
    bp = boxhead;
    while (bp != NULL)
      {
        c = bp->colors;
        while (c != NULL)
          {
            p = c->pixlist;
            while (p != NULL)
              {
/* The following line replaces the next one */
                d[p->offset] = i;
/*
                d[p->y*XSize+p-x] = i;
*/
                p = p->next;
              }
            c = c->next;
          }
        bp = bp->next;
        i++;
      }
 
    /* Compute the new color for each box as the weighted average of all
       colors inside the box */
    bp = boxhead;
    while (bp != NULL)
      {
        i = r = g = b = 0;
        c = bp->colors;
        while (c != NULL)
          {
            r += c->count*(int)(c->r);
            g += c->count*(int)(c->g);
            b += c->count*(int)(c->b);
            i += c->count;
            c = c->next;
          }
        bp->rmin = (unsigned char)(r/i);
        bp->gmin = (unsigned char)(g/i);
        bp->bmin = (unsigned char)(b/i);
        bp = bp->next;
      }

    /* Build new color map */
    memset(rmap,255,256);
    memset(gmap,255,256);
    memset(bmap,255,256);
    rm = rmap;
    gm = gmap;
    bm = bmap;
    bp = boxhead;
    while (bp != NULL)
      {
        *rm = lshift[bp->rmin];              /* Code back to 8 bits */
        *gm = lshift[bp->gmin];
        *bm = lshift[bp->bmin];
        bp = bp->next;
        rm++;gm++;bm++;
      }

    /* Sort the color maps and remap the data */
    sort_cmaps(ncolors);
  }

void ColorImage::free_colors()
  {
    /* Release all malloc()'ed items except the new colormap. */
    struct box *b,*b1;
    int i;

    delete pixels;
    for (i=0; i<cbcount; i++) delete cblock[i];

    b = boxhead;
    while (b != NULL)
      {
        b1 = b->next;
        delete b;
        b = b1;
      }
  }

void ColorImage::sort_cmaps(int nc)
  {
    /* This is a cheapo way to do this and can definately be done better,
       but it is good enough for the 1.0 release....  SRW */
    int i,j,k,jold;
    static unsigned int r[256],g[256],b[256];
    static unsigned int used[256],trans[256];
    unsigned char *d,*rm,*gm,*bm;
    unsigned int v,dark,close;
    int XSize,YSize;

    bzero(used,nc*sizeof(int));

    
    rm = rmap; gm = gmap; bm = bmap;
    for (i=0; i<nc; i++)
      {
        r[i] = *rm;
        g[i] = *gm;
        b[i] = *bm;
        rm++;gm++;bm++;
      }

    /* Find darkest color as the initial entry on the output list */
    j = 0;
    dark = r[0]+g[0]+b[0];
    for (i=1; i<nc; i++)
      {
        v = r[i]+g[i]+b[i];
        if (v < dark)
          {
            j = i;
            dark = v;
          }
      }

    /* Begin filling in the output list */
    trans[j] = 0;
    used[j] = 1;
    jold = j;
    for (i=1; i<nc; i++)
      {
        close = 256*3;
        for (k=0; k<nc; k++)
          {
            if (used[k] == 0)
              {
                v = abs((int)r[jold]-(int)r[k])+
                    abs((int)g[jold]-(int)g[k])+
                    abs((int)b[jold]-(int)b[k]);
                if (v < close)
                  {
                    close = v;
                    j = k;
                  }
              }
          }
        trans[j] = i;
        used[j] = 1;
        jold = j;
      }

    /* Remap the data */
    d = map_plane->GetDataPtr();
    map_plane->GetSize(XSize,YSize);
    for (i=0; i<XSize*YSize; i++)
      {
        *d = trans[*d];
        d++;
      }
    
    /* Install the sorted maps */
    rm = rmap;gm = gmap; bm = bmap;
    for (i=0; i<256; i++)
      {
        *(rm+trans[i]) = r[i];
        *(gm+trans[i]) = g[i];
        *(bm+trans[i]) = b[i];
      }

  }


void copy_ColorImage_header(ColorImage* source,ColorImage* dest)
{
  if (dest->red_plane != NULL) delete dest->red_plane;
  dest->red_plane = source->red_plane;
  if (dest->green_plane != NULL) delete dest->green_plane;
  dest->green_plane = source->green_plane;
  if (dest->blue_plane != NULL) delete dest->blue_plane;
  dest->blue_plane = source->blue_plane;
  if (dest->hue_plane != NULL) delete dest->hue_plane;
  dest->hue_plane = source->hue_plane;
  if (dest->saturation_plane != NULL) delete dest->saturation_plane;
  dest->saturation_plane = source->saturation_plane;
  if (dest->value_plane != NULL) delete dest->value_plane;
  dest->value_plane = source->value_plane;
  if (dest->intensity_plane != NULL) delete dest->intensity_plane;
  dest->intensity_plane = source->intensity_plane;
}

