#include<Token.h>
#include<Register.h>
 
/*********************************************************
* Registry                                               *
*    A registry is a simple device for keeping track of  *
* Tokens. The Register(Token *) function checks to see if*
* a token is in the registry. If not, it adds it to the  *
* registry and returns a unique number assigned to that  *
* token. If the token is already registered, it returns  *
* its (previously assigned) number. Token numbers start  *
* with zero and go up consecutively. The Translate(int)  *
* function is the inverse of Register: it takes a number *
* and returns a pointer to the corresponding token.      *
*                                                        *
*   To iterate down the tokens in a Registry call next(),*
* which will return a pointer to a previously unseen     *
* token, or NULL if all tokens have been seen.           *
*********************************************************/

Registry::Registry (int size)
{
  int i;

  max_size = size;
  current_size = 0;
  marker = 0;
  array = (Token **)malloc(size * sizeof(Token *));
  for (i=0; i < size; i++) 
    *(array + i) = NULL;
}

Registry::~Registry ()
{
  free( array );
  array = NULL;
}

int Registry::find (Token *tok)
{
  int i;
  Token **ptr;

  assert(current_size <= max_size);
  assert(tok != NULL);

  for (i=0, ptr = array; i < current_size; i++, ptr++) {
    if (*ptr == tok) 
      return i;
  }

  return -1;
}

int Registry::Register (Token *tok)
{
  int i;

  i = find(tok);

  if (i >= 0) return i;
  else {
    if (current_size >= max_size) {
      cerr << "Registering out_of_bounds token.\n";
      exit(-1);
    }
    *(array + current_size) = tok;
    current_size++;
    return (1 - current_size);
  }
}

int Registry::Register (Token *tok, int index)
{
  assert(index < max_size);
  assert(index >= 0);
  
  if (NULL != *(array + index)) {
    cerr << "Error: attempt to register token into occupied slot\n";
    exit(-1);
  }
  
  *(array + index) = tok;
  current_size = MAX(current_size, (1 + index));
  return index;
}

Token *Registry::translate (int i)
{
  assert(i >= 0);
  assert(i < max_size);
  assert(array != NULL);

  if (i >= current_size) return NULL;
  else return *(array + i);
}

int Registry::next ()
{
  if (marker >= current_size) {
    return -1;
  }
  else return marker++;
}

void Registry::reset ()
{
  marker = 0;
}

int Registry::count ()
{
  return current_size;
}
