#ifndef TRANSFORM2D_DEFINED
#define TRANSFORM2D_DEFINED
#include<Token.h>
#include<math.h>

/*
 * Axis Direction:
 *    Down == scan order (increasing y means down, increasing x means right)
 *    Up   == anti scan order (increasing y means up, increasing x means left)
 */

enum AxisDirection {Up,Down}; 
enum RotationUnit {Degrees,Radians};
/* enum Boolean {False,True}; */
/* Xlib.h defines Bool, False & True, so we live with their definition ... */
#ifndef Bool
#define Bool int
#define False 0
#define True 1
#endif

class Transform2D : public Token {
  friend Transform2D& operator * (Transform2D& t1, Transform2D& t2);
 private:
  double matrix[3][3];
  Bool affine;       /* TRUE if transformation is affine ([2][0] & [2][1] are 0) */
  Bool translate;    /* FALSE if no translation ([0][2] & [1][2] are 0) */
  Bool rotate;       /* FALSE if no rotation ([0][0] & [1][1] are 1 and
                                                   [0][1] & [1][0] are 0) */
  void SetLogicals();
 public:
  Transform2D();
  Transform2D(double r11, double r12, double tx,
	      double r21, double r22, double ty,
	      double a0 = 0.0,  double a1 = 0.0,  double s = 1.0);
  Transform2D(Transform2D* ts);
  void output(Isrostream&);
  void input(Isristream&);
  char *name() {return "Transform2D";}

  void Rotate(double theta, RotationUnit unit);
  void SetTranslation(double x, double y);
  void Translate(double x, double y);
  int Scale(double factor);
  int SetScale(double factor);
  void Mirror(Bool x_axis, Bool y_axis);
  double Determinant();
  double Minor(int, int);
  Transform2D *Inverse();
  double *Apply(double x, double y, double s = 1.0);
};

Transform2D& operator * (Transform2D& t1, Transform2D& t2);

class CoordSys2D : public Token {
 public:
  CoordSys2D(CoordSys2D& bs);
  CoordSys2D(int xorigin = 0, int yorigin = 0, 
	     AxisDirection xdir = Down, AxisDirection ydir = Down,
	     Transform2D *base_token = NULL, Transform2D *token_base = NULL);
  ~CoordSys2D();

  char *name() {return "CoordSys2D";}
  void CoordSys2D_input(Isristream&);
  void CoordSys2D_output(Isrostream&);
  void input(Isristream& is) {CoordSys2D_input(is);}
  void output(Isrostream& os) {CoordSys2D_output(os);}

  int XOrigin;             /* X offset from upper left of image to Origin (+ or -) */
  int YOrigin;             /* Y offset  "    "     "   "    "   "     "      "     */
  AxisDirection XDir;      /* Down = right, Up = left */
  AxisDirection YDir;      /* Down = down, Up = up    */

  Transform2D *base2token; /* for rotations, etc., of base coordinate system */
  Transform2D *token2base; /*                                                */

  void SetCoordSys2D(int XOrigin, int YOrigin, 
		     AxisDirection XDir, AxisDirection YDir,
		     Transform2D *base_token = NULL,
		     Transform2D *token_base = NULL);
};


Transform2D *overlay(CoordSys2D& b1, CoordSys2D& b2);


#endif /*  TRANSFORM2D_DEFINED */

