#include<TypeDirectory.h>
#include<Token.h>

TypeDirectory* system_table_ptr = NULL;

 TypeDirectory::TypeDirectory(int max_size)
{
  max_table_size = max_size;
  table = (type_directory_element *)malloc(max_size * sizeof(type_directory_element));
  for (int i = 0; i < max_size; i++) {
    table[i].valid = INVALID;
  }
}

 TypeDirectory::~TypeDirectory()
{
  free(table);
}

int TypeDirectory::define(char *name, Token* (*allocator)(), void (*alt_input)())
{
  int index;

  for (index = 0; index < max_table_size; index++) {
    if (table[index].valid == INVALID) break;
    if (0 == strcmp(name, table[index].name)) {
      cerr << "Attempt to redefine token type\n";
      return(-1);
    }
  }

  if (index >= max_table_size) {
    cerr << "Too many token types for table\n";
    return(-1);
  }
    
  table[index].valid = VALID;
  table[index].active = INACTIVE;
  table[index].name = name;
  table[index].allocator = allocator;
  table[index].alt_input = alt_input;
  
  return(index);
}

int TypeDirectory::add_entry(int new_index, char *name, TypeDirectory& old_directory)
{
  int old_index = old_directory.lookup(name);

  if ((new_index >= max_table_size) || (new_index < 0)) {
    cerr << "attempt to add out-of-bounds entry to type_directory\n";
    return(-1);
  }

  if (table[new_index].valid == VALID) {
    cerr << "attempt to add inconsistent entry to a type_directory\n";
    return(-1);
  }
  
  table[new_index].valid = VALID;
  table[new_index].active = INACTIVE;
  table[new_index].name = name;
  table[new_index].allocator = old_directory.allocator(old_index);
  table[new_index].alt_input = old_directory.alt_input(old_index);

  return(new_index);
}

int TypeDirectory::add_entry(int new_index, char *name)
{
  if (system_table_ptr == NULL) ISRinit();

  return add_entry(new_index, name, *system_table_ptr);
}

int TypeDirectory::lookup (char *name)
{
  int i;

  for (i=0; i < max_table_size; i++) {
    if ((table[i].valid == VALID) && (0 == strcmp(name, table[i].name))) 
	return i;
  }

  return -1 ;
}

FN_RET_TOK_PTR *TypeDirectory::allocator(int index)
{
  return(table[index].allocator);
}

FN_RET_VOID *TypeDirectory::alt_input(int index)
{
  return(table[index].alt_input);
}

char *TypeDirectory::name(int index)
{
  return(table[index].name);
}

int TypeDirectory::valid(int index)
{
  return(table[index].valid);
}

void TypeDirectory::deactivate()
{
  int i;

  for(i=0; i < max_table_size; i++) {
    table[i].active = INACTIVE;
  }
}

void TypeDirectory::activate(int i) 
{
  if ((i < 0) || (i >= max_table_size)) {
    cerr << "activating out-of-bounds type index\n";
    exit(-1);
  }
  if (table[i].valid == INVALID) {
    cerr << "activating undefined token type\n";
    exit(-1);
  }
  table[i].active = ACTIVE;
}


int TypeDirectory::active_entries ()
{
  int i;
  int total = 0;
  
  for (i=0; i < max_table_size; i++) {
    if ((table[i].valid == VALID) &&
	(table[i].active == ACTIVE)) {
      total += 1;
    }
  }
  return total;
}

int TypeDirectory::max_active_entry ()
{
  int i;
  
  for (i = max_table_size - 1; i >= 0; i--) {
    if ((table[i].valid == VALID) &&
	(table[i].active == ACTIVE)) {
      return i;
    }
  }
  return 0;
}

void TypeDirectory::print_active_entries (Isrostream& os)
{
  int i;

  for (i=0; i < max_table_size; i++) {
    if ((table[i].valid == VALID) &&
	(table[i].active == ACTIVE)) {
      os << i << " ";
      os.print_string(table[i].name);
      os << "\n";
      if (!os) {
	cerr << "error writing system table\n";
	exit(-1);
      }
    }
  }
  os << "\n";

  return;
}

TypeDirectory* system_type_table()
{
  if (system_table_ptr == NULL) ISRinit();

  return(system_table_ptr);
}
