/*
 * ------------------------------------------------------------------
 * NewPlane.h - Plane tokens
 * Created by Robert Heller on Wed Oct 12 13:03:53 1994
 * ------------------------------------------------------------------
 * Modification History: $Log: NewPlane.h,v $
 * Revision 1.5  1994/11/02  17:52:23  heller
 * Made some additional members protected instead of private.
 *
 * Revision 1.4  1994/11/02  16:55:42  heller
 * Make some members of Plane protected for use by derived Plane classes
 *
 * Revision 1.2  1994/10/12  17:04:14  heller
 * Added RCS keywork hackery
 *
 * ------------------------------------------------------------------
 * Contents:
 * ------------------------------------------------------------------
 *  
 * 
 *            Copyright 1994 University of Massachusetts.
 *                          All rights reserved.
 * 
 *      Permission to copy and modify this software and its documen-
 *      tation only for internal use in your organization is hereby
 *      granted, provided that this notice is retained thereon and
 *      on all copies.  UMASS makes no representations as to the sui-
 *      tability and operability of this software for any purpose.
 *      It is provided "as is" without express or implied warranty.
 * 
 *      UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *      INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 *      NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
 *      INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 *      SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *      IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 *      ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 *      FORMANCE OF THIS SOFTWARE.
 * 
 *      No other rights, including, for example, the right to redis-
 *      tribute this software and its documentation or the right to
 *      prepare derivative works, are granted unless specifically
 *      provided in a separate license agreement.
 * 
 *      Copyright 1994, University of Massachusetts. All rights
 *      reserved.
 * 
 * 
 */
#ifndef _PLANE_
#define _PLANE_

/* assumes -I/users1/vis/kbvision/include */
/* needs -L/users1/vis/kbvision/kbv_ -lvis */

#include <Token.h>
#include <X11/Intrinsic.h>
#include <Transform2D.h>
#ifdef KHOROS
#include<viff.h>
#endif
#include <ctype.h>
#include <math.h>
#ifdef HUGE_VAL
#undef HUGE_VAL
#endif

#ifdef KBV
extern "C" {
#include <KBVUtil.h>	// this is probably broken.  As usual  
/* silly (non-standard) KBV defines */
#undef malloc
#undef calloc
#undef realloc
#undef free
#undef strlen
#undef strcmp
#undef strncpy
#undef putenv
#undef sleep
#undef TRUE
#undef FALSE
#include <vstd.h>
#include <Iaf.h>
#include <Tcb.h>
}
#endif

#include <limits.h>
#ifdef HUGE_VAL
#undef HUGE_VAL
#endif

#define UNINITIALIZED -1

const int Undefined = INT_MIN;
class BytePlane;
class IntPlane;
class FloatPlane;

class Plane : public Token {
#ifdef KBV
  friend Plane *kbv2llvs(IMAGE_PTR img);
#endif
  friend class BytePlane;
  friend class IntPlane;
  friend class FloatPlane;
  friend Boolean Incr_PlaneBitDrawWorkCallback(XtPointer clientData);
 protected:
  int	XSize, YSize;		/* 2D size of plane */
  int	Level;			/* relative resolution */
  int pixel_size;               /* bytes per pixel */
  int data_size;                /* (xsize * ysize * pixel_size) */
  int plane_type;               /* see llvs_plane.h */

  unsigned char **Data2D;       /* pixel as 2D array */

  XImage *ximage;		/* display image */
  double minval, maxval;	/* data range */
  const MAXGREY = 128;		/* max number of grey levels (7-bits) */
  Boolean DontScale;		/* supress scaling */
  Pixel colorMapping[MAXGREY];	/* pixel mapping */
  /* Incrmental state info */
  XtIntervalId Incr_workId;	/* work id */
  int   Incr_Currow, Incr_Curcol; /* current row for incremental update */
  int Incr_maskrow, Incr_maskcol;
  int Incr_hpixsize, Incr_vpixsize;
  Boolean Incr_refresh;
  double Incr_intensity_scale;
  Colormap Incr_cmap;
  Boolean Incr_dithering;
  Display *Incr_display;
  Drawable Incr_drawable;
  GC Incr_gc;
  struct {int isrcx,isrcy,idestx,idesty,imw, imh;} Incr_PutImParams;
  Boolean Incr_drawGreyImageCallback();
  Boolean Incr_ditherBitImageCallback();
  const int XIMThresh = 256;
  void  plane_min_max();
  void FreeData();
  void Init (int xsize, int ysize, int xorigin, int yorigin, 
	     AxisDirection xdir, AxisDirection ydir, char *str);

  void plane_output(Isrostream& stream);
  void	readbin(Isristream& stream);
  void	readfmt(Isristream& stream);

 private:
  unsigned char* read_Data1D(Isristream& instream, int plane_type, 
			     int *xsize, int *ysize,
			     int *xloc, int *yloc, int *level,
			     AxisDirection *xdir, AxisDirection *ydir);

  friend BytePlane *read_bit_plane_as_byte_plane(Isristream& is);
  friend IntPlane *read_short_plane_as_int_plane(Isristream& is);

  void FreeData2D();
  void SetData2D();
  void InitData();
  void write_plane_header(Isrostream& stream);
  int	writebin(Isrostream& stream);
  void	writefmt(Isrostream& stream);
  void  drawGreyImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans);
  void  ditherBitImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans);
  void  Incr_drawGreyImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans);
  void  Incr_ditherBitImage(Display *d,int XIMWidth,int XIMHeight,Colormap cmap,Transform2D* trans);
    

public:
  CoordSys2D coord;
  CoordSys2D *coordinates() { return &coord; }
  char    Name[16];               /* Plane name */
  unsigned char *Data1D;	/* pixel data as a flat vector */

  Plane();
  ~Plane ();

  void output(Isrostream&);
  void input(Isristream&);

  int GetXOrigin() {return coord.XOrigin;}
  int GetYOrigin() {return coord.YOrigin;}
  AxisDirection GetXDir() {return coord.XDir;}
  AxisDirection GetYDir() {return coord.YDir;}
  void GetSize(int &xsize,int &ysize) {xsize = XSize;ysize = YSize;}
  int GetPixelSize() { return pixel_size; }
  int GetType() { return plane_type; }
  void SetCoordinates(int xorigin,int yorigin,
		      AxisDirection xdir,AxisDirection ydir);
  void GetMinMax(int &minx,int &maxx,int &miny,int &maxy);
  void GetMin(int &minx,int &miny);
  void GetMax(int &maxx,int &maxy);
  unsigned char *GetDataPtr() { return Data1D; }
  void GetMinMaxVal(double &min,double &max)
  	{if (minval == maxval && maxval == 0.0) plane_min_max();
  	 min = minval; max = maxval;}
#ifdef KHOROS
  xvimage* viff();
#endif
#ifdef KBV
  IMAGE_PTR kbv();
#endif
  virtual void BitDraw(Display* display, Drawable drawable, GC gc, Transform2D* trans,
		    float xmin, float ymin, float xmax, float ymax,Visual *visual,
		    unsigned int depth,Colormap cmap);
  void SetColorMap(unsigned char rmap[],unsigned char gmap[],
		   unsigned char bmap[],int nc,Display *display,Colormap cmap);
};

class BytePlane : public Plane {
 public:
  inline unsigned char& pixel(int x,int y)
    { if (coord.XDir == Up) x = -x;
      if (coord.YDir == Up) y = -y;
      return(Data2D[y][x]); }
  void output(Isrostream& stream);
  void input(Isristream& stream);
  char *name() { return "BytePlane"; }
  
  BytePlane ();
  BytePlane (Plane& pl, char *str = "");
  BytePlane (int XSize, int YSize, int XOrigin, int YOrigin, 
	     AxisDirection XDir = Up, AxisDirection YDir = Up, char *Name = "");
  ~BytePlane();
#ifdef TIFFLIB
  friend void writeTiffFile(char *filename,BytePlane *bpl);
  friend Token* ReadTiffFile(char *filename);
#endif
};


BytePlane *read_byte_plane(Isristream&);


class IntPlane : public Plane {
 public:
  inline long int& pixel(int x,int y)
    { if (coord.XDir == Up) x = -x;
      if (coord.YDir == Up) y = -y;
      return( ((long int **)Data2D)[y][x]); }
  // Constructor & Destructor
  IntPlane (int XSize, int YSize, int XOrigin, int YOrigin,
	      AxisDirection XDir=Up, AxisDirection YDir=Up, char *Name = "");
  IntPlane ();
  IntPlane(Plane &Pl, char *str = "");
  ~IntPlane();

  void output(Isrostream& stream);
  void input(Isristream& stream);
  char *name() { return "IntPlane"; }
};

IntPlane *read_int_plane(Isristream&);

class FloatPlane : public Plane {
 public:
  inline float& pixel(int x,int y)
	{ if (coord.XDir == Up) x = -x;
	  if (coord.YDir == Up) y = -y;
	  return( ((float **)Data2D)[y][x]); }
   // Constructor & Destructor
   FloatPlane (int XSize, int YSize, int XOrigin,int YOrigin,
	       AxisDirection XDir=Up, AxisDirection YDir=Up,
	       char *Name = "");
  FloatPlane ();
  FloatPlane(Plane &pl, char *str = "");
  ~FloatPlane();

  void output(Isrostream& stream);
  void input(Isristream& stream);
  char *name() { return "FloatPlane"; }
};

FloatPlane *read_float_plane(Isristream&);
void copy_plane_header(Plane *source, Plane *dest);
#endif

