#include<assert.h>
#include<Token.h>
#include<Register.h>
dnan Token::NaN_Value;
dnan Token::PosInf_Value;
dnan Token::NegInf_Value;
int Token::token_count = 0;
Registry* print_registry = NULL;

Token::Token()
{
  if (token_count == 0)
  {
  	NaN_Value.nan_parts.fraction_low = 0;
  	NaN_Value.nan_parts.bits = 0;
  	NaN_Value.nan_parts.qnan_bit = 1;
  	NaN_Value.nan_parts.exponent = 0x7ff;
  	NaN_Value.nan_parts.sign = 0;

  	PosInf_Value.inf_parts.fraction_low = 0;
  	PosInf_Value.inf_parts.bits = 0;
  	PosInf_Value.inf_parts.exponent = 0x7ff;
  	PosInf_Value.inf_parts.sign = 0;

  	NegInf_Value.inf_parts.fraction_low = 0;
  	NegInf_Value.inf_parts.bits = 0;
  	NegInf_Value.inf_parts.exponent = 0x7ff;
  	NegInf_Value.inf_parts.sign = 1;
  }
  token_type = -1;
  token_count++;
}

Token::~Token()
{
  token_count--;
}

void Token::output(Isrostream& dest)
{
  cerr << "Output function not defined\n";
  exit(-1);
}

void Token::input(Isristream& source)
{
  cerr << "Input function not defined\n";
  exit(-1);
}

char *Token::name()
{
  cerr << "Name function not defined\n";
  exit(-1);
  return NULL;
}

int Token::type_index ()
{
  TypeDirectory *type_table = system_type_table();
    
  if (token_type > type_table->max_possible_entry()) {
    cerr << "Error: internal state of token_type munged\n";
    exit(-1);
  }

  if (token_type < 0) {
    token_type = type_table->lookup(name());
  }
  return(token_type);
}

void Token::Draw(Display* display, Drawable drawable, 
			 GC gc, Transform2D* trans,
			 float xmin, float ymin, float xmax, float ymax)
{
  return;
}


void Token::BitDraw(Display* display, Drawable drawable, 
			 GC gc, Transform2D* trans,
			 float xmin, float ymin, float xmax, float ymax,
			 Visual *visual,
			 unsigned int depth,Colormap cmap)
{
  return;
}


void Token::Transform(Transform2D *trans)
{
  cerr << "Undefined Transform function.\n";
  exit(-1);
}


void read_token_number_and_type (Isristream& is, ISR_FILE_TOKEN_PTR& token_ptr)
{
  char *tokentype_name;
  short int short_token_type_ptr;

  is >> "<token #" >> token_ptr.token_index >> "type " >> short_token_type_ptr;
  token_ptr.tokentype_index = (int)short_token_type_ptr;

  /* check for errors */
  if ((!is) || (token_ptr.token_index < -1)) {
    cerr << "Token index error (probably out of bounds)\n";
    exit(-3);
  }
  else if ((token_ptr.tokentype_index < -1) || 
	   (token_ptr.tokentype_index > is.tokentype_table->table_size())) {
    cerr << "Token type index out of bounds\n";
    exit(-3);
  }
  
  /* if this type has not been seen before, copy it from system table into
     input stream table, otherwise just verify that the name is correct. */

  if (token_ptr.token_index == -1) tokentype_name = "NULL";
  else tokentype_name = is.tokentype_table->name(token_ptr.tokentype_index);

  if (NULL == tokentype_name) {
    int temp;

    is >> tokentype_name >> ">";
    temp = is.tokentype_table->add_entry(token_ptr.tokentype_index, 
					tokentype_name);
    if ((!is) || (temp < 0)) {
      cerr << "unknown token type " << tokentype_name << " in file\n";
      exit(-4);
    }
  }
  else {
    is >> (const char *)tokentype_name >> ">";
    if (!is) {
      cerr << "Tokentype " << tokentype_name << " does not match tokentype_index "
	<< token_ptr.tokentype_index << "\n";
      exit(-5);
    }
  }
  return;
}

Token* read_token_file_ptr (Isristream& is, Token *preallocated_token/* = NULL*/)
{
  ISR_FILE_TOKEN_PTR file_ptr;
  Token *token_ptr;
  FN_RET_TOK_PTR *alloc_function;

  assert(print_registry != NULL);

  read_token_number_and_type(is, file_ptr);

  /* check for null pointer, and return NULL if so. */
  if ((file_ptr.token_index == -1) && (file_ptr.tokentype_index == -1))
    return NULL;

  /* if index is negative, then a read error occured. This is
     not the same as reading a pointer to a previously unencountered
     token. This means the file format was wrong, and the pointer is
     garbage. The best you can do is return the NULL pointer. */

  if (file_ptr.token_index < 0) {
    cerr << "error: unable to read token pointer (token index out of range)\n";
    exit(-1);
  }

  token_ptr = print_registry->translate(file_ptr.token_index);

  /* if token has not been allocated, use & register
     default token supplied by the user; otherwise
     use the previous token and delete the user's */

  if (token_ptr == NULL) {
    if (preallocated_token != NULL) {
      token_ptr = preallocated_token;
      print_registry->Register(token_ptr, file_ptr.token_index);
    }
    else {
      alloc_function = is.tokentype_table->allocator(file_ptr.tokentype_index);
      if (alloc_function != NULL) {
	token_ptr = (*alloc_function)();
	print_registry->Register(token_ptr, file_ptr.token_index);
      }
      else {
	cerr << "reading pointer to unknown token type.\n";
	return NULL;
      }
    }
  }
  else if (preallocated_token != NULL) {
    /* if the token has already been allocated and stored in the
       registry, but we are trying to put it into a preallocated 
       spot, then we have a problem. */
    cerr << "preallocated token incompatible with previous allocations.\n";
    return NULL;
  }

  return token_ptr;
}

Token* read_token_ptr ( Isristream& is)
{ 
  return read_token_file_ptr(is);
}


/*
Isrostream& operator << (Isrostream& os, Token& token)
{
  return (os << &token);
}
*/

Isrostream& operator << (Isrostream& os, Token *token)
{
  if (token == NULL) write_token_number_and_type(os, NULL, -1);
  else token->print_token_ptr(os);

  return os;
}

void Token::print_token_ptr (Isrostream& os)
{
  int index = -1;

  if (print_registry == NULL) {
    cerr << "Error: No print registry for recursive token printing\n";
    exit(-1);
  }

  index = print_registry->find(this);
  if (index < 0) {
    cerr << "Error: printing unregistered token. Trace method probably did not access it.\n";
      exit(-1);
  }

  write_token_number_and_type(os, this, index);
}

/* This function writes an abstract pointer to a token into stream (file) os.
   The form of the pointer is <token #N, type TYPE TYPENAME> in ASCII or
   simply N TYPE in binary. */
void write_token_number_and_type(Isrostream& os, Token *token, int index)
{
  char *print_name;
  short int type_index;

  if (token == NULL) {
    print_name = "NULL";
    type_index = -1;
    index = -1;
  }
  else {
    print_name = token->name();
    type_index = (short int)token->type_index();
  }

  os << "<token #" << index << " type " << type_index << " " << print_name << ">";
  if (!os) {
    cerr << "Unable to write token # " << index << "to output stream\n";
    exit(-1);
  }
}

int Register(Token *tok)
{
  if (print_registry == NULL) return -1;
  if (tok == NULL) return -1;
  return print_registry->Register(tok);
}

