#ifndef TOKEN_DEFINED
#define TOKEN_DEFINED

#include<string.h>
#include<Xlib.h>
#include<Isrstream.h>
#include<TypeDirectory.h>
#include<ISRinit.h>
#include <math.h>
#define _IEEE 1	// missing???
#include <nan.h>
// ScumOS's (and Linux's) /usr/include/nan.h is brain dead...
#if (defined(sun) && !defined(__svr4__)) || defined(i386) 
typedef union
{
         struct	
	 {
#ifdef ISR3_LITTLEENDIAN
	    unsigned fraction_low:32;
            unsigned bits:20;
	    unsigned exponent :11;
	    unsigned sign     : 1;
#else
	    unsigned sign     : 1;
	    unsigned exponent :11;
            unsigned bits:20;
	    unsigned fraction_low:32;
#endif
         } inf_parts;
	 struct 
	 {
#ifdef ISR3_LITTLEENDIAN
	    unsigned fraction_low: 32;
	    unsigned bits     :19;
	    unsigned qnan_bit : 1;
            unsigned exponent :11;
	    unsigned sign     : 1;
#else
	    unsigned sign     : 1;
            unsigned exponent :11;
	    unsigned qnan_bit : 1;
	    unsigned bits     :19;
	    unsigned fraction_low: 32;
#endif
         } nan_parts;
         double d;

} dnan;
#endif
#ifndef IsNANorINF
#define IsNANorINF(X)  (((dnan *)&(X))->nan_parts.exponent == 0x7ff)
#endif
#ifndef IsINF
#define IsINF(X)  (((dnan *)&(X))->inf_parts.bits == 0 &&  \
                    ((dnan *)&(X))->inf_parts.fraction_low == 0)
#endif

#ifndef IsPosNAN
#define IsPosNAN(X)  (((dnan *)&(X))->nan_parts.sign == 0)
#endif
#ifndef IsNegNAN
#define IsNegNAN(X)  (((dnan *)&(X))->nan_parts.sign == 1)
#endif
#ifndef NaN
#define NaN(X)  (((dnan *)&(X))->nan_parts.exponent == 0x7ff)
#endif

#define INDENT 3
#define MAXSTRINGLENGTH 80

/* A structure for internal use only that records both the
   token number and the tokentype number of a token pointer
   in a file. */

typedef struct ISR_file_token_ptr {
  int token_index;
  int tokentype_index;
} ISR_FILE_TOKEN_PTR;

/***********************************************************
* class Token            11/8/93            Bruce A. Draper*
*                                                          *
*   Token is the top-level class in the ISR3.1 class hier- *
* archy, and is not directly instantiable. The Token class *
* is meant to address the one issue all ISR objects share: *
* I/O. The functions in this class handle file-based I/O   *
* (both ASCII and binary), and will probably be used for   *
* interprocess communication as well.                      *
*                                                          *
*                                                          *
*   To define a new Token subclass, a user must provide    *
* three (virtual) functions. Two of the functions are I/O  *
* primitives: input and output. These functions take an    *
* Isristream (input) or Isrostream (output) and write to   *
* it just as they would an istream or ostream, i.e. by     *
* using << or >>. See Isrstream.h for details about        *
* about Isrstreams. The output function should print out   *
* the fields of a token in an easily readable format. If   *
* the token contains pointers to other tokens, the write   *
* function should write those pointers just like any other *
* field (by using <<), and the read routine should read    *
* them using >>. The Isrstream will take care of the       *
* address translation, and insure against infinite loops.  *
*                                                          *
*   What allows the system to avoid infinite loops and     *
* translate old addresses to new ones is a user-defined    *
* function called trace(). For each new token subclass,    *
* a trace function must be defined that calls the (system- *
* defined) function Register on every Token pointed to by  *
* the original Token. This allows the system to trace down *
* a data tree, avoiding infinite loops and setting up      *
* address translation tables. It is also used by the system*
* for the function free(), which recursively deletes all   *
* the Tokens in a data tree.                               *
*                                                          *
***********************************************************/

class TokenField {
 public:
  char* name;	       /* Name of field */
  int   offset;	       /* Offset of field */
  int   size;	       /* size of field (only used for POINTER) */
  /* types -- (see page 5-4 of the KBV Prog. ref) */
  enum FieldType {INTEGER,       /* 32-bit signed integer - KBV_INTEGER_DT */
	          FLOAT,         /* 32-bit floating point - KBV_FLOAT_DT */
	          STRING,        /* null terminated string - KBV_STRING_DT */
	          /* I doubt we will support these, but here they are anyway. */
	          CONSTELLATION, /* 2D array of bits (region mask) 
                                   - KBV_CONSTELLATION_DT */
	          FPOLYGON,      /* an ordered set of FP coords - KBV_FPOLYGON_DT */
	          LINE_SEGMENT,  /* an ordered pair of FP coords - KBV_LINE_SEGMENT_DT */
	          FPOINT,	/* a FP coord pair - KBV_FPOINT_DT */
	          FRECTANGLE,    /* a pair of FPOINTs - KBV_FRECTANGLE_DT */
	          TOKENSUBSET,   /* a list of token specifiers and a pointer to a Tokenset
				     KBV_TOKENSUBSET_DT */
	          POINTER,	/* an arbitary block of memory - KBV_POINTER_DT */
	          FCHAIN,        /* an order set of FP coords - KBV_FCHAIN_DT */
	          CONIC_SPLINE,  /* an ordered set of FP coords - KBV_CONIC_SPLINE_DT */
                  /* The following datatypes apparently have no KBV correlate */
		  TOKEN_PTR,     /* pointer to a token */
		  DOUBLE         /* 64-bit floating point */
	        };

  FieldType typecode;  /* type of field */
};

class Isristream;
class Isrostream;
class Transform2D;
class CoordSys2D;

class Token {

  friend class ISROStream;
  friend class ISRIStream;

/* friend Isrostream& operator<< ( Isrostream&, Token&); */
  friend Isrostream& operator<< ( Isrostream&, Token*);
  friend Isristream& operator>> ( Isristream&, Token&);

 public:
  Token();
  virtual ~Token();

  int type_index ();
  void print_token_ptr(Isrostream&);
  void demand (Token* ptr);
  Token *translate(Token *old_addr);

  /* required for I/O capability */
  virtual void output(Isrostream&);
  virtual void input(Isristream&);
  virtual void trace() { return; }
  virtual char* name();
  /* required for TKS compatibility & ISRAccess/ISRPut */
  virtual int field_count() { return 0; }
  virtual TokenField *fields() { return NULL; }
  /* required for graphics capability */
  virtual CoordSys2D *coordinates() { return NULL; }
  virtual void Transform(Transform2D *trans);
  virtual void Draw(Display* display, Drawable drawable, GC gc, Transform2D* trans,
		    float xmin, float ymin, float xmax, float ymax);
  virtual void BitDraw(Display* display, Drawable drawable, GC gc, Transform2D* trans,
		    float xmin, float ymin, float xmax, float ymax,
		    Visual *visual,
		    unsigned int depth,Colormap cmap);
  /* put rasterize() here */

  static dnan NaN_Value;
  static dnan PosInf_Value;
  static dnan NegInf_Value;
 private:
  int token_type;
  static int token_count;
};

Isrostream& operator<< ( Isrostream&, Token&);
Isrostream& operator<< ( Isrostream&, Token*);

Token* read_token_ptr(Isristream&);
int Register(Token *tok);

void write_token_number_and_type (Isrostream&, Token *, int);
void read_token_number_and_type (Isristream&, ISR_FILE_TOKEN_PTR&);
Token* read_token_file_ptr(Isristream& is, Token *preallocated_token = NULL);
Token* read_top_level_token_ptr(Isristream& is, Token *preallocated = NULL);
#endif /* TOKEN_DEFINED */
