/*
 * ------------------------------------------------------------------
 * TokenStream.h - TokenStream defs
 * Created by Bruce Draper
 * $Id: TokenStream.h,v 1.3 1994/10/17 18:08:00 heller Exp $
 * ------------------------------------------------------------------
 * Modification History: $Log: TokenStream.h,v $
 * Revision 1.3  1994/10/17  18:08:00  heller
 * Added support for writing viff 3-band color images.
 *
 * Revision 1.2  1994/10/14  19:45:43  heller
 * Added support for reading viff 3-band color images.
 *
 * ------------------------------------------------------------------
 * Contents:
 * ------------------------------------------------------------------
 *  
 * 
 *            Copyright 1994 University of Massachusetts.
 *                          All rights reserved.
 * 
 *      Permission to copy and modify this software and its documen-
 *      tation only for internal use in your organization is hereby
 *      granted, provided that this notice is retained thereon and
 *      on all copies.  UMASS makes no representations as to the sui-
 *      tability and operability of this software for any purpose.
 *      It is provided "as is" without express or implied warranty.
 * 
 *      UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 *      INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 *      NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
 *      INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 *      SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 *      IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 *      ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 *      FORMANCE OF THIS SOFTWARE.
 * 
 *      No other rights, including, for example, the right to redis-
 *      tribute this software and its documentation or the right to
 *      prepare derivative works, are granted unless specifically
 *      provided in a separate license agreement.
 * 
 *      Copyright 1994, University of Massachusetts. All rights
 *      reserved.
 * 
 * 
 */
#ifndef TOKENSTREAM_DEFINED
#define TOKENSTREAM_DEFINED
  
#include <stddef.h>
#include <iostream.h>
#include <fstream.h>
#include <iomanip.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <Isrstream.h>
#include <TypeDirectory.h>
#include <Register.h>
#include <ISRinit.h>
#include <Token.h>
enum StreamFormat {untyped, isra, isrb, plane, viff, im, tks, isr2, ps, gif,
		   tiff};

class Token;

/* Generic Declaration for Format Streams */

class IFormatStream {
 public:
  virtual void open(char *) = 0;
  virtual void close() = 0;
  virtual void read_token(Token&) = 0;
  virtual Token* read_token_ptr() = 0;
  virtual int operator ! () {return 1;}
};

class OFormatStream {
 public:
  virtual void open(char *) = 0;
  virtual void close() = 0;
  virtual void write_token(Token&) = 0;
  virtual int operator ! () {return 1;}
};

/* Particular Format Streams */

class ISROStream : public OFormatStream {
 public:
  ISROStream(StreamFormat);
  virtual ~ISROStream();

  void open(char *);
  void close();
  void write_token(Token&);
  int operator ! () {return !(*isr_stream);}

 protected:
  Isrostream *isr_stream;
};

class ISRIStream : public IFormatStream {
 public:
  ISRIStream(StreamFormat);
  virtual ~ISRIStream();

  void open(char *);  
  void close();
  void read_token(Token&);
  Token* read_token_ptr();
  int operator ! () {return !(*isr_stream);}

 protected:
  Isristream *isr_stream;
  Token* internal_read_token_ptr (Token* preallocated = NULL);
};

class Plane;
extern Token* read_llvs_plane(Isristream& is);
extern void copy_plane_header(Plane *, Plane *);
class ColorImage;
extern void copy_ColorImage_header(ColorImage *,ColorImage *);

class PlaneOStream : public ISROStream {
 public:
  PlaneOStream() : ISROStream(isrb) {}
  virtual ~PlaneOStream();

  void write_token(Token&);
};
  
class PlaneIStream : public ISRIStream {
 public:
  PlaneIStream() : ISRIStream(isrb) {}
  virtual ~PlaneIStream();

  void read_token(Token&);
  Token* read_token_ptr();
};

#ifdef KHOROS
extern "C" struct xvimage *readimage(char *filename);
extern Plane *viff2llvs(xvimage *image);
extern ColorImage *viff2ColorImage(xvimage *image);
extern void writeplaneimage(char *filename, Plane *pl);
extern int writecolorimage(char *filename, ColorImage *ci);
extern "C" void freeimage(struct xvimage *image);

class ViffOStream : public OFormatStream {
 public:
  ViffOStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void write_token(Token&);
  /* Note: stream must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};
  
class ViffIStream : public IFormatStream {
 public:
  ViffIStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void read_token(Token&);
  Token* read_token_ptr();
  /* Note: streams must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};
#endif

#include <ColorImage.h>
extern void writeGifFile(char *filename,ColorImage *cimg);
ColorImage* GIF2ColorImage(char *fname);

class GifOStream : public OFormatStream {
 public:
   GifOStream() {filename = NULL;}
   void open(char *);
   void close() { return; }
   void write_token(Token&);
   virtual int operator ! () {return 0;}
 private:
   char *filename;
};

class GifIStream : public IFormatStream {
 public:
  GifIStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void read_token(Token&);
  Token* read_token_ptr();
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};

#ifdef TIFFLIB

class BytePlane;
extern void writeTiffFile(char *filename,BytePlane *bpl);
extern void writeTiffFile(char *filename,ColorImage *cimg);
extern Token* ReadTiffFile(char *filename);

class TiffOStream : public OFormatStream {
 public:
   TiffOStream() {filename = NULL;}
   void open(char *);
   void close() { return; }
   void write_token(Token&);
   virtual int operator ! () {return 0;}
 private:
   char *filename;
};

class TiffIStream : public IFormatStream {
 public:
  TiffIStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void read_token(Token&);
  Token* read_token_ptr();
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};
#endif

#ifdef KBV
extern void writekbvimage(char *filename, Plane *pl);
extern Token *readkbvplane(char *filename);

class ImOStream : public OFormatStream {
 public:
  ImOStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void write_token(Token&);
  /* Note: stream must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};
  
class ImIStream : public IFormatStream {
 public:
  ImIStream() { filename = NULL; }

  void open(char *);
  void close() { return; }
  void read_token(Token&);
  Token* read_token_ptr();
  /* Note: streams must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return 0;}

 private:
  char *filename;
};

class TksOStream : public OFormatStream {
 public:
  TksOStream() { filename = NULL; tksname = NULL;}
  virtual ~TksOStream() {close();}

  void open(char *fname);
  void close() {if (tksname != NULL) delete tksname;}
  void write_token(Token&);
  /* Note: stream must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return 0;}

 private:
  char *filename;
  char *tksname;
};
  
class TksIStream : public IFormatStream {
 public:
  TksIStream() { filename = NULL; tksname = NULL;}
  virtual ~TksIStream() {close();}
  
  void open(char *fname);
  void close();
  void read_token(Token&);
  Token* read_token_ptr();
  /* Note: streams must be OK, or else Khoros would have exited with error. */
  virtual int operator ! () {return(tksname == NULL);}

 private:
  char *filename;
  char *tksname;
};

#endif

class TokenStream {
 public:
  char *name;                       /* filename or stream name */
 
  int openp();
  int MatchFormatToFilename(char *filename);
  virtual int SetFormat(StreamFormat);

 protected:
  StreamStatus state;               /* Open or Closed */
  StreamFormat format;              /* ISR, PLANE, VIFF ... */

  void TokenStreamInit();
  void TokenStreamFree();
};

class TokenOStream : public TokenStream {
  friend TokenOStream& operator << (TokenOStream&, Token&);
  friend TokenOStream& operator << (TokenOStream&, Token*);
  friend int operator ! (TokenOStream&);

 public:
  int SetFormat(StreamFormat);

  void open(char *);
  void close();

  TokenOStream();
  TokenOStream(StreamFormat);
  TokenOStream(char *);
  virtual ~TokenOStream();

 private:
  OFormatStream *format_stream;
};  

TokenOStream& operator << (TokenOStream&, Token&);
TokenOStream& operator << (TokenOStream&, Token*);
int operator ! (TokenOStream&);

class TokenIStream : public TokenStream { 
  friend TokenIStream& operator >> (TokenIStream&, Token&); 
  friend int operator ! (TokenIStream&);

 public: 
  int SetFormat(StreamFormat);

  void open(char *);
  void close();
 
  TokenIStream();
  TokenIStream(StreamFormat);
  TokenIStream(char *);
  virtual ~TokenIStream();

  Token* read_token_ptr();

 private:
  IFormatStream *format_stream;
};

int operator ! (TokenIStream&);

#endif /* TOKENSTREAM_DEFINED */

