#include <Transform2D.h>

#ifndef PI
#define PI M_PI
#endif

CoordSys2D::CoordSys2D(CoordSys2D& bs)
{
  XOrigin = bs.XOrigin;
  YOrigin = bs.YOrigin;
  XDir = bs.XDir;
  YDir = bs.YDir;
  base2token = bs.base2token;
  token2base = bs.token2base;
}

CoordSys2D::CoordSys2D(int xorigin/* = 0*/, int yorigin/* = 0*/, 
		       AxisDirection xdir/* = Down*/, AxisDirection ydir/* = Down*/,
		       Transform2D *base_token/* = NULL*/, Transform2D *token_base/* = NULL*/)
{
  XOrigin = xorigin;
  YOrigin = yorigin;
  XDir = xdir;
  YDir = ydir;
  base2token = base_token;
  token2base = token_base;
}

CoordSys2D::~CoordSys2D()
{
  if (base2token != NULL) {
    free(base2token);
    base2token = NULL;
  }
  if (token2base != NULL) {
    free(token2base);
    token2base = NULL;
  }
}

void write_axis_direction(Isrostream& os, AxisDirection dir)
{
  if (os.format == ASCII) {
    if (dir == Up) os << "Up";
    else os << "Down";
  }
  else if (os.format == BINARY) {
    if (dir == Up) os << 'U';
    else os << 'D';
  }
  else {
    cerr << "Error writing AxisDirection\n";
    return;
  }
}

void CoordSys2D::CoordSys2D_output(Isrostream& os)
{
  os << "CoordSys2D:: XOrigin " << XOrigin << " YOrigin " << YOrigin;
  os << " XDir ";
  write_axis_direction(os, XDir);
  os << " YDir ";
  write_axis_direction(os, YDir);
  os << "\nbase2token " << base2token << " token2base " << token2base << "\n";
}

AxisDirection read_axis_direction(Isristream& is)
{
  if (is.format == ASCII) {
    char ch = is.get();
    if (ch == 'U') {
      is >> "p";
      return Up;
    }
    else if (ch == 'D') {
      is >> "own";
      return Down;
    }
    else {
      cerr << "Error reading (ASCII) AxisDirection\n";
      return Up;
    }
  }
  else {
    char ch = is.get();
    if (ch == 'U') return Up;
    else if (ch == 'D') return Down;
    else {
      cerr << "Error reading (BINARY) AxisDirection\n";
      return Up;
    }
  }
}


void CoordSys2D::CoordSys2D_input(Isristream& is)
{
  is >> "CoordSys2D:: XOrigin" >> XOrigin >> "YOrigin" >> YOrigin;
  is >> "XDir ";
  XDir = read_axis_direction(is);
  is >> "YDir ";
  YDir = read_axis_direction(is);
  is >> "base2token";
  base2token = (Transform2D *)read_token_ptr(is);
  is >> "token2base";
  token2base = (Transform2D *)read_token_ptr(is);
}

void CoordSys2D::SetCoordSys2D(int XOrigin, int YOrigin, 
			       AxisDirection XDir, AxisDirection YDir,
			       Transform2D *base_token/* = NULL*/,
			       Transform2D *token_base/* = NULL*/)
{
  XOrigin = XOrigin;
  YOrigin = YOrigin;
  XDir = XDir;
  YDir = YDir;
  base2token = base_token;
  token2base = token_base;
}

Transform2D::Transform2D()
{
  matrix[0][0] = 1.0;
  matrix[0][1] = 0.0;
  matrix[0][2] = 0.0;
  matrix[1][0] = 0.0;
  matrix[1][1] = 1.0;
  matrix[1][2] = 0.0;
  matrix[2][0] = 0.0;
  matrix[2][1] = 0.0;
  matrix[2][2] = 1.0;
  affine = True;
  translate = False;
  rotate = False;
}

Transform2D::Transform2D(Transform2D* ts)
{
	for (int i=0;i < 3;i++)
	{
		for (int j=0;j < 3;j++)
		{
			matrix[i][j] = ts->matrix[i][j];
		}
	}
	SetLogicals();
}

Transform2D::Transform2D(double r11, double r12, double tx,
			 double r21, double r22, double ty,
			 double a0/* = 0.0*/,  double a1/* = 0.0*/,  double s/* = 1.0*/)
{
  matrix[0][0] = r11;
  matrix[0][1] = r12;
  matrix[0][2] = tx;
  matrix[1][0] = r21;
  matrix[1][1] = r22;
  matrix[1][2] = ty;
  matrix[2][0] = a0;
  matrix[2][1] = a1;
  matrix[2][2] = s;

  SetLogicals();
}

void Transform2D::output(Isrostream& os) 
{
  os << "Transform2D::\n";
  os << matrix[0][0] << matrix[0][1] << matrix[0][2] << "\n";
  os << matrix[1][0] << matrix[1][1] << matrix[1][2] << "\n";
  os << matrix[2][0] << matrix[2][1] << matrix[2][2] << "\n";
}

void Transform2D::input(Isristream& is) 
{
  is >> "Transform2D::\n";
  is >> matrix[0][0] >> matrix[0][1] >> matrix[0][2];
  is >> matrix[1][0] >> matrix[1][1] >> matrix[1][2];
  is >> matrix[2][0] >> matrix[2][1] >> matrix[2][2];
}

void Transform2D::SetLogicals()
{
  if ((matrix[2][0] == 0.0) && (matrix[2][1] == 0.0)) affine = True;
  else affine = False;

  if ((matrix[0][2] == 0.0) && (matrix[1][2] == 0.0)) translate = False;
  else translate = True;

  if ((matrix[0][1] == 0.0) && (matrix[1][0] == 0.0)) rotate = False;
  else rotate = True;

  if ((matrix[2][2] != 0.0) && (matrix[2][2] != 1.0)) {
    matrix[0][0] /= matrix[2][2];
    matrix[1][1] /= matrix[2][2];
    if (rotate == True) {
      matrix[0][1] /= matrix[2][2];
      matrix[1][0] /= matrix[2][2];
    }
    if (translate == True) {
      matrix[0][2] /= matrix[2][2];
      matrix[1][2] /= matrix[2][2];
    }
    if (affine == False) {
      matrix[2][0] /= matrix[2][2];
      matrix[2][1] /= matrix[2][2];
    }
    matrix[2][2] = 1.0;
  }
}

Transform2D& operator * (Transform2D& t1, Transform2D& t2)
{
  Transform2D *new_t;

  new_t = new Transform2D;
  
  /* a = a1a2 + b1d2 + c1g2 */
  new_t->matrix[0][0] = t1.matrix[0][0] * t2.matrix[0][0];
  if ((t1.rotate == True) && (t2.rotate == True))
    new_t->matrix[0][0] += t1.matrix[0][1] * t2.matrix[1][0];
  if ((t2.affine == False) && (t2.translate == True))
    new_t->matrix[0][0] += t1.matrix[0][2] * t2.matrix[2][0];

  /* b = a1b2 + b1e2 + c1h2 */
  if (t2.rotate == True) new_t->matrix[0][1] = t1.matrix[0][0] * t2.matrix[0][1];
  if (t1.rotate == True) new_t->matrix[0][1] += t1.matrix[0][1] * t2.matrix[1][1];
  if ((t2.affine == False) && (t1.translate == True))
    new_t->matrix[0][1] += t1.matrix[0][2] * t2.matrix[1][2];

  /* c = a1c2 + b1f2 + c1i2 */
  if (t2.translate == True) {
    new_t->matrix[0][2] = t1.matrix[0][0] * t2.matrix[0][2];
    if (t1.rotate == True) new_t->matrix[0][2] += t1.matrix[0][1] * t2.matrix[1][2];
  }
  if (t1.translate == True) new_t->matrix[0][2] += t1.matrix[0][2]; /* matrix[2][2] */

  /* d = d1a2 + e1d2 + f1g2 */
  if (t1.rotate == True) new_t->matrix[1][0] = t1.matrix[1][0] * t2.matrix[0][0];
  if (t2.rotate == True) new_t->matrix[1][0] += t1.matrix[1][1] * t2.matrix[1][0];
  if ((t2.affine == False) && (t1.translate == True))
    new_t->matrix[1][0] += t1.matrix[1][2] * t2.matrix[2][0];
  
  /* e = d1b2 + e1e2 + f1h2 */
  new_t->matrix[1][1] = t1.matrix[1][1] * t2.matrix[1][1];
  if ((t1.rotate == True) && (t2.rotate == True))
    new_t->matrix[1][1] += t1.matrix[1][0] * t2.matrix[0][1];
  if ((t2.affine == False) && (t1.translate == True)) 
    new_t->matrix[1][1] += t1.matrix[1][2] * t2.matrix[2][1];

  /* f = d1c2 + e1f2 + f1i2 */
  if (t2.translate == True) {
    new_t->matrix[1][2] = t1.matrix[1][1] * t2.matrix[1][2];
    if (t1.rotate == True) new_t->matrix[1][2] += t1.matrix[1][0] * t2.matrix[0][2];
  }
  if (t1.translate == True) new_t->matrix[1][2] += t1.matrix[1][2]; /* t2.matrix[2][2]*/

  /* g = g1a2 + h1d2 + i1g2 */
  if (t1.affine == False) {
    new_t->matrix[2][0] = t1.matrix[2][0] * t2.matrix[0][0];
    if (t2.rotate == True) new_t->matrix[2][0] += t1.matrix[2][1] * t2.matrix[1][0];
  }
  if (t2.affine == False) new_t->matrix[2][0] += t2.matrix[2][0]; /* t1.matrix[2][2] */

  /* h = g1b2 + h1e2 + i1h2 */
  if (t1.affine == False) {
    new_t->matrix[2][1] = t1.matrix[2][1] * t2.matrix[1][1];
    if (t2.rotate == True) new_t->matrix[2][1] += t1.matrix[2][0] * t2.matrix[0][1];
  }
  if (t2.affine == False) new_t->matrix[2][1] += t2.matrix[2][1]; /*t1.matrix[2][2] */

  /* i = g1c2 + h1f2 + i1i2 */
  /* new_t->matrix[2][2] = t1.matrix[2][2] * t2.matrix[2][2]; */
  if ((t1.affine == False) && (t2.translate == True)) {
    new_t->matrix[2][2] += t1.matrix[2][0] * t2.matrix[0][2];
    new_t->matrix[2][2] += t1.matrix[2][1] * t2.matrix[1][2];
  }

  new_t->SetLogicals();

  return *new_t;
}

double *Transform2D::Apply(double x, double y, double s/* = 1.0*/)
{
  double *pt;

  pt = new double[3];
  
  //cerr << "*** Transform2D::Apply(" << x << "," << y << "," << s << ") => ";

  pt[0] = matrix[0][0] * x;
  if (rotate == True) pt[0] += matrix[0][1] * y;
  if (translate == True) pt[0] += matrix[0][2] * s;

  pt[1] = matrix[1][1] * y;
  if (rotate == True) pt[1] += matrix [1][0] * x;
  if (translate == True) pt[1] += matrix[1][2] * s;

  pt[2] = matrix[2][2] * s;
  if (affine == False) {
    pt[2] += matrix[2][0] * x;
    pt[2] += matrix[2][1] * y;
  }

  if (pt[2] == 0.0)
  {
  	//cerr << pt[0] << "," << pt[1] << "," << pt[2] << "\n";
	return pt;
  }

  if (pt[2] != 1.0) {
    pt[0] /= pt[2];
    pt[1] /= pt[2];
    pt[2] = 1.0;
  }

  //cerr << pt[0] << "," << pt[1] << "," << pt[2] << "\n";
  return pt;
}

int Transform2D::Scale (double factor)
{
  if (factor == 0.0) return -1;
  
  matrix[2][2] /= factor;
  SetLogicals();
  return 1;
}

int Transform2D::SetScale (double factor)
{
  if (factor == 0.0) return -1;
  
  //cerr << "*** Transform2D::SetScale(): factor = " << factor << "\n";
  //cerr << "matrix[0][0] (before) = " << matrix[0][0] << "\n";
  //cerr << "matrix[1][1] (before) = " << matrix[1][1] << "\n";
  matrix[2][2] = 1.0/factor;
  SetLogicals();
  //cerr << "matrix[0][0] (after) = " << matrix[0][0] << "\n";
  //cerr << "matrix[1][1] (after) = " << matrix[1][1] << "\n";
  return 1;
}

void Transform2D::Mirror(Bool x_axis, Bool y_axis)
{
  if (x_axis == True) {
    matrix[0][0] = - matrix[0][0];
    matrix[0][1] = - matrix[0][1];
    matrix[0][2] = - matrix[0][2];
  }
  if (y_axis == True) {
    matrix[1][0] = - matrix[1][0];
    matrix[1][1] = - matrix[1][1];
    matrix[1][2] = - matrix[1][2];
  }
}

void Transform2D::Translate(double x, double y)
{
  matrix[0][2] += x * matrix[2][2];
  matrix[1][2] += y * matrix[2][2];
  
  if (affine == False) {
    matrix[0][0] += x * matrix[2][0];
    matrix[0][1] += x * matrix[2][1];
    matrix[1][0] += y * matrix[2][0];
    matrix[1][1] += y * matrix[2][1];
  }

  SetLogicals();
}
    
void Transform2D::SetTranslation(double x, double y)
{
  matrix[0][2] = x * matrix[2][2];
  matrix[1][2] = y * matrix[2][2];
  
  if (affine == False) {
    matrix[0][0] += x * matrix[2][0];
    matrix[0][1] += x * matrix[2][1];
    matrix[1][0] += y * matrix[2][0];
    matrix[1][1] += y * matrix[2][1];
  }

  SetLogicals();
}
    
void Transform2D::Rotate(double theta, RotationUnit unit)
{
  double angle, c, s;

  /* convert to radians, if necessary */
  if (unit == Degrees) angle = (PI * theta) / 180.0;
  else angle = theta;

  c = cos(angle);
  s = sin(angle);

  matrix[0][0] = c * matrix[0][0] + s * matrix[1][0];
  matrix[0][1] = c * matrix[0][1] + s * matrix[1][1];
  matrix[0][2] = c * matrix[0][2] + s * matrix[1][2];
  matrix[1][0] = c * matrix[1][0] - s * matrix[0][0];
  matrix[1][1] = c * matrix[1][1] - s * matrix[0][1];
  matrix[1][2] = c * matrix[1][2] - s * matrix[0][2];

  SetLogicals();
}

double Transform2D::Determinant()
{
  /* remember: matrix[2][2] = 1.0, and for affines [2][0] and [2][1] are 0.0 */
  if (affine = True) return ((matrix[0][0] * matrix[1][1]) - 
			     (matrix[0][1] * matrix[1][0] * matrix [2][2]));
  else return ((matrix[0][0] * matrix[1][1]) +
	       (matrix[0][1] * matrix[1][2] * matrix [2][0]) +
	       (matrix[0][2] * matrix[1][0] * matrix [2][1]) -
	       (matrix[0][2] * matrix[1][1] * matrix [2][0]) -
	       (matrix[0][1] * matrix[1][0]) -
	       (matrix[0][0] * matrix[1][2] * matrix [2][1]));
}

double Transform2D::Minor(int row, int col)
{
  double TwoByTwo[4];
  int ctr = 0, r, c;

  for (r=0; r < 3; r++) {
    if (r != row) {
      for (c=0; c < 3; c++) {
	if (c != col) 
	  TwoByTwo[ctr++] = matrix[r][c];
      }
    }
  }
  
  return ((TwoByTwo[0] * TwoByTwo[3]) - (TwoByTwo[1] * TwoByTwo[2]));
}

Transform2D *Transform2D::Inverse()
{
  Transform2D *new_trans = new Transform2D;
  int i, j;
  float temp;
  double det = Determinant();

  if (det == 0.0) return NULL;

  for (i=0; i < 3; i++) {
    for (j=0; j < 3; j++) {
      new_trans->matrix[i][j] = (Minor(j,i) / det);
      temp = (i + j) / 2.0;
      if (temp != floor(temp)) new_trans->matrix[i][j] = -new_trans->matrix[i][j];
    }
  }

  new_trans->SetLogicals();
  return new_trans;
}
  

Transform2D *overlay(CoordSys2D& b1, CoordSys2D& b2)
{
  Transform2D *trans;

  trans = new Transform2D;

  if (b1.XDir == Up) trans->Mirror(True,False);
  if (b1.YDir == Up) trans->Mirror(False,True);

  trans->Translate(b1.XOrigin, b1.YOrigin);
  trans->Translate(-b2.XOrigin, -b2.YOrigin);

  if (b2.XDir == Up) trans->Mirror(True,False);
  if (b2.YDir == Up) trans->Mirror(False,True);

  /* code for base2token token2base should be added here */

  return trans;
}

