/*
 *----------------------------------------------------------------------
 *
 * Copyright 1991-1994, University of Massachusetts.  All rights reserved.
 * Permission to copy and modify this software and its documen-
 * tation only for internal use in your organization is hereby
 * granted, provided that this notice is retained thereon and
 * on all copies.  UMASS makes no representations as to the sui-
 * tability and operability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 * 
 * UMASS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
 * NESS.  IN NO EVENT SHALL UMASS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY OTHER DAMAGES WHAT-
 * SOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PER-
 * FORMANCE OF THIS SOFTWARE.
 * 
 * No other rights, including, for example, the right to redis-
 * tribute this software and its documentation or the right to
 * prepare derivative works, are granted unless specifically
 * provided in a separate license agreement.
 *---------------------------------------------------------------------
 * $Log: viff.cc,v $
// Revision 1.4  1994/10/17  18:08:19  heller
// Added support for writing viff 3-band color images.
//
// Revision 1.3  1994/10/14  19:46:28  heller
// Added support for reading viff 3-band color images.
//
// Revision 1.2  1994/10/14  13:55:41  heller
// Added support for all datatypes supported by viff files.
//
 */

/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 >>>>
 >>>>         File Name: lllvs2viff.c
 >>>>
 >>>>      Program Name: llvs2viff
 >>>>
 >>>> Date Last Updated: Tue May 11 13:34:50 1993 
 >>>>
 >>>>          Routines: lllvs2viff - the library call for llvs2viff
 >>>>
 >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>   <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<*/


/* $Id: viff.cc,v 1.4 1994/10/17 18:08:19 heller Exp $ */
static char rcsid[] = "$Id: viff.cc,v 1.4 1994/10/17 18:08:19 heller Exp $";


#include <ViffPlane.h>


/****************************************************************
*    
*    

****************************************************************/

xvimage *Plane::viff()
{
    struct xvimage *img = new xvimage;
    unsigned long int data_storage_type;

    switch (plane_type) {
        case LLVS_PLF_BYTE: data_storage_type = VFF_TYP_1_BYTE; break;
        case LLVS_PLF_INT: data_storage_type = VFF_TYP_4_BYTE; break;
        case LLVS_PLF_FLOAT: data_storage_type = VFF_TYP_FLOAT; break;
    }

    img = createimage (YSize,XSize,data_storage_type,1,1,NULL,
                       0,0,VFF_MS_NONE,VFF_MAPTYP_NONE,
                       VFF_LOC_IMPLICIT,0);
    if (NULL == img) return NULL;

    img->startx = coord.XOrigin;
    img->starty = coord.YOrigin;
    memcpy((unsigned char *) img->imagedata,Data1D,(YSize*XSize*pixel_size));
    return img;
}

xvimage *ColorImage::viff()
{
    struct xvimage *img = new xvimage;
    unsigned long int data_storage_type;

    if (red_plane != NULL && green_plane != NULL && blue_plane != NULL)
    {
    	data_storage_type = VFF_TYP_1_BYTE;
    	int YSize,XSize;
    	red_plane->GetSize(XSize,YSize);
    	img = createimage (YSize,XSize,data_storage_type,1,3,NULL,
    			   0,0,VFF_MS_NONE,VFF_MAPTYP_NONE,
			   VFF_LOC_IMPLICIT,0);
	if (img == NULL) return(NULL);
	img->color_space_model = VFF_CM_genericRGB;
	img->startx = red_plane->GetXOrigin();
	img->starty = red_plane->GetYOrigin();
	int pixel_size = red_plane->GetPixelSize();
	int plsize = (YSize*XSize*pixel_size);
	memcpy((unsigned char*)(img->imagedata),
	       (unsigned char*)(red_plane->GetDataPtr()),
	       plsize);
	memcpy((unsigned char*)(img->imagedata)+plsize,
	       (unsigned char*)(green_plane->GetDataPtr()),
	       plsize);
	memcpy((unsigned char*)(img->imagedata)+plsize+plsize,
	       (unsigned char*)(blue_plane->GetDataPtr()),
	       plsize);
	return(img);
    } else if (hue_plane != NULL &&
	       saturation_plane != NULL && 
	       value_plane != NULL)
    {
    	data_storage_type = VFF_TYP_FLOAT;
    	int YSize,XSize;
    	hue_plane->GetSize(XSize,YSize);
    	img = createimage (YSize,XSize,data_storage_type,1,3,NULL,
    			   0,0,VFF_MS_NONE,VFF_MAPTYP_NONE,
			   VFF_LOC_IMPLICIT,0);
	if (img == NULL) return(NULL);
	img->color_space_model = VFF_CM_HSV;
	img->startx = hue_plane->GetXOrigin();
	img->starty = hue_plane->GetYOrigin();
	int pixel_size = hue_plane->GetPixelSize();
	int plsize = (YSize*XSize*pixel_size);
	memcpy((unsigned char*)(img->imagedata),
	       (unsigned char*)(hue_plane->GetDataPtr()),
	       plsize);
	memcpy((unsigned char*)(img->imagedata)+plsize,
	       (unsigned char*)(saturation_plane->GetDataPtr()),
	       plsize);
	memcpy((unsigned char*)(img->imagedata)+plsize+plsize,
	       (unsigned char*)(value_plane->GetDataPtr()),
	       plsize);
	return(img);
    } else return(NULL);
}


Plane *viff2llvs(xvimage *image)
{
  int data_storage_type;
  Plane *pl;
  int xorigin, yorigin;
  int plsize;
  int xx,yy;
  
  if (image->starty < 0) yorigin = 0;
  else yorigin = image->starty;
  if (image->startx < 0) xorigin = 0;
  else xorigin = image->startx;
  
  switch (image->data_storage_type) {
  case VFF_TYP_1_BYTE: 
    data_storage_type = LLVS_PLF_BYTE; 
    pl = new BytePlane(image->row_size, image->col_size,
		       xorigin, yorigin, Down, Down, NULL);
    plsize = image->row_size * image->col_size;
    break; 
  case VFF_TYP_4_BYTE: data_storage_type = LLVS_PLF_INT; 
    pl = new IntPlane(image->row_size, image->col_size,
		      xorigin, yorigin, Down, Down, NULL);
    plsize = image->row_size * image->col_size * sizeof(long int);
    break;
  case VFF_TYP_FLOAT: data_storage_type = LLVS_PLF_FLOAT; 
    pl = new FloatPlane(image->row_size, image->col_size,
			xorigin, yorigin, Down, Down, NULL);
    plsize = image->row_size * image->col_size * sizeof(float);
    break;
  case VFF_TYP_BIT: data_storage_type = LLVS_PLF_BYTE;
    pl = new BytePlane(image->row_size, image->col_size,
		       xorigin, yorigin, Down, Down, NULL);
    plsize = -1;
    break;
  case VFF_TYP_2_BYTE: data_storage_type = LLVS_PLF_INT;
    pl = new IntPlane(image->row_size, image->col_size,
                            xorigin, yorigin, Down, Down, NULL);
    plsize = -1;
    break;
  case VFF_TYP_DCOMPLEX:  
  case VFF_TYP_COMPLEX:  
  case VFF_TYP_DOUBLE: data_storage_type = LLVS_PLF_FLOAT;
    pl = new FloatPlane(image->row_size, image->col_size,
    			xorigin, yorigin, Down, Down, NULL);
    plsize = -1;
    break;
  }
  
  if (plsize != -1)
  {
    memcpy((unsigned char*)(pl->GetDataPtr()),(unsigned char *) image->imagedata,plsize);
  } else
  {
    switch (image->data_storage_type)
    {
      case VFF_TYP_BIT:
      {
        unsigned char *cptr = pl->GetDataPtr();
        unsigned char *bptr = (unsigned char *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
      	  int k = 0;
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    int bit = 1 << k;
	    if (((*bptr) & bit) != 0)
	    {
	      *cptr++ = 1;
	    } else {
	      *cptr++ = 0;
	    }
	    if (k == 7)
	    {
	      k = 0;
	      bptr++;
	    } else k++;
	  }
	  if (k != 0) bptr++;
        }
      }
      break;
      case VFF_TYP_2_BYTE:
      {
        long int *cptr = (long int*) pl->GetDataPtr();
        short int *bptr = (short int *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_DOUBLE:
      {
        float *cptr = (float*) pl->GetDataPtr();
        double *bptr = (double *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_COMPLEX:
      {
        float *cptr = (float*) pl->GetDataPtr();
        float *bptr = (float *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	    bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_DCOMPLEX:
      {
        float *cptr = (float*) pl->GetDataPtr();
        double *bptr = (double *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	    bptr++;
	  }
        }
      }
      break;
    }
  }
  return(pl);
}

void writeplaneimage(char *filename, Plane *pl)
{
  writeimage(filename, pl->viff());
}

int writecolorimage(char *filename, ColorImage *ci)
{
  xvimage *viff = ci->viff();
  if (viff == NULL) return(0);
  else
  {
    writeimage(filename, viff);
    return(1);
  }
}


ColorImage* viff2ColorImage(struct xvimage *image)
{
  int data_storage_type;
  ColorImage *ci;
  Plane *p1, *p2, *p3;
  int xorigin, yorigin;
  int plsize;
  int xx,yy;
  
  if (image->starty < 0) yorigin = 0;
  else yorigin = image->starty;
  if (image->startx < 0) xorigin = 0;
  else xorigin = image->startx;

  ci = new ColorImage;
  switch (image->data_storage_type) {

  case VFF_TYP_BIT: 
  case VFF_TYP_1_BYTE: 
    data_storage_type = LLVS_PLF_BYTE; 
    p1 = new BytePlane(image->row_size, image->col_size,
		       xorigin, yorigin, Down, Down, NULL);
    p2 = new BytePlane(image->row_size, image->col_size,
		       xorigin, yorigin, Down, Down, NULL);
    p3 = new BytePlane(image->row_size, image->col_size,
		       xorigin, yorigin, Down, Down, NULL);
    if (image->data_storage_type == VFF_TYP_BIT) plsize = -1;
    else plsize = image->row_size * image->col_size;
    break; 
  case VFF_TYP_2_BYTE:
  case VFF_TYP_4_BYTE: data_storage_type = LLVS_PLF_INT; 
    p1 = new IntPlane(image->row_size, image->col_size,
		      xorigin, yorigin, Down, Down, NULL);
    p2 = new IntPlane(image->row_size, image->col_size,
		      xorigin, yorigin, Down, Down, NULL);
    p3 = new IntPlane(image->row_size, image->col_size,
		      xorigin, yorigin, Down, Down, NULL);
    if (image->data_storage_type == VFF_TYP_2_BYTE) plsize = -1;
    else plsize = image->row_size * image->col_size * sizeof(long int);
    break;
  case VFF_TYP_DCOMPLEX:
  case VFF_TYP_COMPLEX:
  case VFF_TYP_DOUBLE:
  case VFF_TYP_FLOAT: data_storage_type = LLVS_PLF_FLOAT; 
    p1 = new FloatPlane(image->row_size, image->col_size,
			xorigin, yorigin, Down, Down, NULL);
    p2 = new FloatPlane(image->row_size, image->col_size,
			xorigin, yorigin, Down, Down, NULL);
    p3 = new FloatPlane(image->row_size, image->col_size,
			xorigin, yorigin, Down, Down, NULL);
    if (image->data_storage_type == VFF_TYP_FLOAT)
    	plsize = image->row_size * image->col_size * sizeof(float);
    else plsize = -1;
    break;
  }
  if (image->color_space_model == VFF_CM_HSV &&
      data_storage_type == LLVS_PLF_FLOAT)
  {
    ci->hue_plane = (FloatPlane*) p1;
    ci->saturation_plane = (FloatPlane*) p2;
    ci->value_plane = (FloatPlane*) p3;
  } else if (data_storage_type == LLVS_PLF_BYTE)
  {
    ci->red_plane = (BytePlane*) p1;
    ci->green_plane = (BytePlane*) p2;
    ci->blue_plane = (BytePlane*) p3;
  } else
  {
    cerr << "Don't know how to deal with this color viff image.\n";
    delete ci;
    delete p1;
    delete p2;
    delete p3;
    return NULL;
  }

  if (plsize != -1)
  {
     memcpy((unsigned char*)(p1->GetDataPtr()),
	    (unsigned char *) image->imagedata,
	    plsize);
     memcpy((unsigned char*)(p2->GetDataPtr()),
	    ((unsigned char *) image->imagedata)+plsize,
	    plsize);
     memcpy((unsigned char*)(p3->GetDataPtr()),
	    ((unsigned char *) image->imagedata)+plsize+plsize,
	    plsize);
  } else
  {
    switch (image->data_storage_type)
    {
      case VFF_TYP_BIT:
      {
      	unsigned char *cptr = p1->GetDataPtr();
      	unsigned char *bptr = (unsigned char *)image->imagedata;
      	for (yy = 0; yy < image->col_size;yy++)
        {
      	  int k = 0;
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    int bit = 1 << k;
	    if (((*bptr) & bit) != 0)
	    {
	      *cptr++ = 1;
	    } else {
	      *cptr++ = 0;
	    }
	    if (k == 7)
	    {
	      k = 0;
	      bptr++;
	    } else k++;
	  }
	  if (k != 0) bptr++;
        }
        cptr = p2->GetDataPtr();
      	for (yy = 0; yy < image->col_size;yy++)
        {
      	  int k = 0;
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    int bit = 1 << k;
	    if (((*bptr) & bit) != 0)
	    {
	      *cptr++ = 1;
	    } else {
	      *cptr++ = 0;
	    }
	    if (k == 7)
	    {
	      k = 0;
	      bptr++;
	    } else k++;
	  }
	  if (k != 0) bptr++;
        }
        cptr = p3->GetDataPtr();
      	for (yy = 0; yy < image->col_size;yy++)
        {
      	  int k = 0;
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    int bit = 1 << k;
	    if (((*bptr) & bit) != 0)
	    {
	      *cptr++ = 1;
	    } else {
	      *cptr++ = 0;
	    }
	    if (k == 7)
	    {
	      k = 0;
	      bptr++;
	    } else k++;
	  }
	  if (k != 0) bptr++;
        }
      }
      break;
      case VFF_TYP_2_BYTE:
      {
        long int *cptr = (long int*) p1->GetDataPtr();
        short int *bptr = (short int *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
	cptr = (long int*) p2->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
	cptr = (long int*) p3->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_DOUBLE:
      {
        float *cptr = (float*) p1->GetDataPtr();
        double *bptr = (double *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
        cptr = (float*) p2->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
        cptr = (float*) p3->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_COMPLEX:
      {
        float *cptr = (float*) p1->GetDataPtr();
        float *bptr = (float *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	    bptr++;
	  }
        }
        cptr = (float*) p2->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
        cptr = (float*) p3->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
      case VFF_TYP_DCOMPLEX:
      {
        float *cptr = (float*) p1->GetDataPtr();
        double *bptr = (double *)image->imagedata;
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	    bptr++;
	  }
        }
        cptr = (float*) p2->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
        cptr = (float*) p3->GetDataPtr();
        for (yy = 0; yy < image->col_size;yy++)
        {
	  for (xx = 0; xx < image->row_size;xx++)
	  {
	    *cptr++ = *bptr++;
	  }
        }
      }
      break;
    }
  }
  return (ci);
}
